
import tramway.helper.roi as helper
from tramway.helper.roi import IgnoredInputWarning, UnitRegions, GroupedRegions

"""
This transition module wraps some functionalities from the :mod:`tramway.helper.roi` module.
"""

def add_metadata(self, analysis, pkg_version=[]):
    return helper.Helper().add_metadata(analysis, pkg_version)

class Collections(helper.RoiCollections):
    def __init__(self, group_overlapping_roi=False, metadata=add_metadata, verbose=False):
        helper.RoiCollections.__init__(self, group_overlapping_roi, metadata=metadata, verbose=verbose)
        self._numeric_format = helper.RoiCollections.numeric_format.fget(self)
    @property
    def numeric_format(self):
        return self._numeric_format
    @numeric_format.setter
    def numeric_format(self, fmt):
        if isinstance(fmt, int):
            fmt = '{{:0>{:d}d}}'.format(fmt)
        self._numeric_format = fmt
    def bounding_boxes(self, sr_index):
        rs = self.regions
        units = rs.region_to_units(sr_index)
        bb = {}
        for coll in units:
            if isinstance(rs, GroupedRegions):
                i, j = rs.collection_range(coll)
                bbs = rs.unit_region[i:j]
            else:
                bbs = rs.unit_region[coll]
            bb[coll] = [ bbs[u] for u in units[coll] ]
        if (isinstance(rs, GroupedRegions) and len(rs.index) == 1) or \
                (isinstance(rs, UnitRegions) and len(rs.unit_region) == 1):
            bb = bb[coll]
        return bb




# -*- coding: utf-8 -*-

# Copyright © 2017-2018, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from ..base import *
from ..grid import RegularMesh
from tramway.core.scaler import *
from math import *
import numpy as np
import pandas as pd
import scipy.sparse as sparse
from scipy.cluster.vq import kmeans, kmeans2
from scipy.spatial.distance import cdist
from collections import OrderedDict


class KMeansMesh(Voronoi):
    """K-Means and Voronoi based tessellation.

    Attributes:

        avg_probability (float): probability of a point to be in a given cell (controls the
            number of cells and indirectly their size).

    Other Attributes:

        _min_distance (float): scaled minimum distance between adjacent cell centers;
            *not used*.
    """
    def __init__(self, scaler=None, min_probability=None, avg_probability=None, \
        min_distance=None, initial='grid', **kwargs):
        Voronoi.__init__(self, scaler)
        #self.min_probability = min_probability
        #self.max_probability = None
        self.avg_probability = avg_probability
        #self.local_probability = None
        self._min_distance = min_distance
        self.initial = initial

    def _preprocess(self, points, **kwargs):
        init = self.scaler.init
        points = Voronoi._preprocess(self, points)
        if init and self._min_distance is not None:
            self._min_distance = self.scaler.scale_distance(self._min_distance)
        if self.initial == 'grid':
            try:
                avg_distance = kwargs['avg_distance']
            except KeyError:
                avg_distance = None
            else:
                avg_distance = self.scaler.scale_distance(avg_distance)
            grid = RegularMesh(avg_probability=self.avg_probability,
                min_distance=self._min_distance, avg_distance=avg_distance)
            grid.tessellate(points)
            self._cell_centers = grid._cell_centers
            #self.lower_bound = grid.lower_bound
            #self.upper_bound = grid.upper_bound
        elif self.initial == 'random':
            lower_bound = np.asarray(points).min(axis=0, keepdims=True)
            upper_bound = np.asarray(points).max(axis=0, keepdims=True)
            if self.avg_probability:
                n_cells = int(round(1. / self.avg_probability))
            else:
                raise ValueError('avg_probability (or avg_location_count) not defined')
            self._cell_centers = np.random.rand(n_cells, points.shape[1])
            self._cell_centers = self._cell_centers * (upper_bound - lower_bound) + lower_bound
        elif self.initial == 'center':
            initial_spread = kwargs.pop('initial_spread', 1e-2)
            points = np.asarray(points)
            lower_bound = points.min(axis=0, keepdims=True)
            upper_bound = points.max(axis=0, keepdims=True)
            center = points.mean(axis=0, keepdims=True)
            if self.avg_probability:
                n_cells = int(round(1. / self.avg_probability))
            else:
                raise ValueError('avg_probability (or avg_location_count) not defined')
            self._cell_centers = np.random.randn(n_cells, points.shape[1])
            self._cell_centers = self._cell_centers * (initial_spread * (upper_bound - lower_bound)) + center
        self.roi_subset_size = 10000
        self.roi_subset_count = 10
        return points

    def tessellate(self, points, tol=1e-6, prune=2.5, plot=False, **kwargs):
        """Grow the tessellation.

        Attributes:
            points: see :meth:`~tramway.tessellation.base.Tessellation.tessellate`.
            tol (float): error tolerance.
                Passed as `thresh` to :func:`scipy.cluster.vq.kmeans`.
            prune (bool or float): prunes the Voronoi and removes the edges which length
                is greater than `prune` times the median edge length;
                ``True`` is translated to the default value.
        """
        points = self._preprocess(points, **kwargs)
        self._cell_centers, _ = kmeans(np.asarray(points), self._cell_centers, \
            thresh=tol)

        if prune: # inter-center-distance-based pruning
            if prune is True: # backward compatibility
                prune = 2.5 # 2.5 is empirical
            self._postprocess(adjacency_label=True)
            A = sparse.tril(self.cell_adjacency, format='coo')
            i, j, k = A.row, A.col, A.data
            if self._adjacency_label is None:
                if np.max(k) == 1:
                    k = np.arange(k.size)
                    self._cell_adjacency = sparse.csr_matrix((np.tile(k, 2),
                        (np.r_[i, j], np.r_[j, i])), shape=A.shape)
                self._adjacency_label = np.ones(k.size, dtype=bool)
            else:
                l = 0 < self._adjacency_label[k]
                i, j, k = i[l], j[l], k[l]
            x = self._cell_centers
            d = x[i] - x[j]
            d = np.sum(d * d, axis=1) # square distance
            d0 = np.median(d)
            edge = k[d0 * prune < d] # edges to be discarded
            if edge.size:
                self._adjacency_label[edge] = False



def _metric(knn=None, **kwargs):
    if isinstance(knn, (tuple, list)):
        knn = knn[0]
    if knn is None:
        return None
    else:
        return 'euclidean'

setup = {
    'make': KMeansMesh,
    'make_arguments': OrderedDict((
        ('min_distance', ()),
        ('avg_distance', ()),
        ('avg_probability', ()),
        ('avg_location_count', dict(args=('-c', '--location-count'), kwargs=dict(type=int, default=80, help='average number of locations per cell'), translate=True)),
        ('metric', dict(parse=_metric)),
        )),
    }

__all__ = ['KMeansMesh', 'setup']




# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from ..attribute import AnalyzerNode, first
from ..artefact import Analysis
import warnings
from tramway.tessellation.base import Partition, Voronoi
from tramway.tessellation.time import TimeLattice


class Mpl(AnalyzerNode):
    """
    Matplotlib plotting utilities for 2D data; no time support.
    """
    __slots__ = ()
    def plot(self, tessellation, locations=None, axes=None,
            voronoi_options=dict(), delaunay_options=None,
            location_options=dict(), **kwargs):
        """
        Plot the particle locations and the Voronoi graph
        (or tessellation).

        `plot` does not access any `RWAnalyzer` attribute
        and can be safely called with no side effect.
        """
        from tramway.helper.tessellation import cell_plot
        try:
            voronoi_options = dict(centroid_style=None, color='rrrr') | voronoi_options
            location_options = dict(alpha=.2, color='k') | location_options
            kwargs = dict(aspect='equal', show=False) | kwargs
        except TypeError: # Python < 3.9
            voronoi_options, _options = dict(centroid_style=None, color='rrrr'), voronoi_options
            voronoi_options.update(_options)
            location_options, _options = dict(alpha=.2, color='k'), location_options
            location_options.update(_options)
            kwargs, _kwargs = dict(aspect='equal', show=False), kwargs
            kwargs.update(_kwargs)
        if axes is not None:
            kwargs['axes'] = axes
        if isinstance(tessellation, Analysis):
            tessellation = tessellation.data
        if isinstance(tessellation, Partition):
            sampling = tessellation
            tessellation = sampling.tessellation
            if locations is None:
                locations = sampling.points
            else:
                sampling = None
        else:
            sampling = None
        if isinstance(tessellation, TimeLattice):
            if tessellation.spatial_mesh is None:
                raise TypeError('no tessellation found; only time segments')
            tessellation = tessellation.spatial_mesh
        if isinstance(tessellation, Voronoi):
            if sampling is None:
                sampling = Partition(locations, tessellation)
        else:
            raise TypeError('tessellation type not supported: {}'.format(type(tessellation)))
        return cell_plot(sampling, voronoi=voronoi_options, locations=location_options,
                delaunay=delaunay_options, **kwargs)

    def animate(self, fig, sampling, axes=None,
            voronoi_options=dict(), location_options=dict(), **kwargs):
        """
        As this method is of limited interest, it has been poorly tested.

        The `RWAnalyzer.time` attribute is accessed.
        """
        from matplotlib import animation
        from tramway.plot import mesh as tplt
        if axes is None:
            axes = fig.gca()
        #
        if isinstance(sampling, Analysis):
            sampling = sampling.data
        nsegments = self._eldest_parent.time.n_time_segments(sampling)
        # copied/pasted from mapper.mpl.Mpl.animate
        anim_kwargs = dict(blit=True, cache_frame_data=False, save_count=nsegments,
                repeat=False, interval=600)
        more_kwargs = dict(repeat_delay=None, fargs=None)
        more_kwargs.update(anim_kwargs)
        for kw in more_kwargs:
            try:
                arg = kwargs.pop(kw)
            except KeyError:
                pass
            else:
                if kw == 'interval' and arg in ('rt', 'realtime', 'real-time'):
                    arg = self._eldest_parent.time.window_shift * 1e3
                anim_kwargs[kw] = arg
        #
        _iter = self._eldest_parent.time.as_time_segments
        #
        try:
            voronoi_options = dict(centroid_style=None, color='rrrr') | voronoi_options
            location_options = dict(alpha=.2, color='k') | location_options
        except TypeError: # Python < 3.9
            voronoi_options, _options = dict(centroid_style=None, color='rrrr'), voronoi_options
            voronoi_options.update(_options)
            location_options, _options = dict(alpha=.2, color='k'), location_options
            location_options.update(_options)
        location_options['markersize'] = location_options.pop('size', 3) # for plot
        #location_options['s'] = location_options.pop('size', 8) # for scatter
        #
        first_segment = first(_iter(sampling, return_times=False))
        x, y = [ sampling.points[col].values for col in 'xy' ]
        glyphs, = axes.plot(x, y, '.', **location_options)
        tplt.plot_voronoi(first_segment, axes=axes, **voronoi_options)
        axes.set_aspect(kwargs.get('aspect', 'equal'))
        def init():
            return glyphs,
        def draw_segment(sampling):
            if isinstance(sampling, tuple):
                times, sampling = sampling
            else:
                times = None
            x, y = [ sampling.points[col].values for col in 'xy' ]
            glyphs.set_data(x, y)
            #xy = sampling.points[list('xy')].values
            #glyphs.set_array(xy)
            return glyphs,
        return animation.FuncAnimation(fig, draw_segment, init_func=init,
                frames=_iter(sampling, return_times=True), **anim_kwargs)

    def plot_cell_indices(self, tessellation, axes=None,
            voronoi_options=dict(), delaunay_options=None,
            aspect='equal', title=None, **kwargs):
        """
        Plot the Voronoi graph and Voronoi cell indices.

        The placement of the text elements is not optimized.
        """
        from tramway.plot.mesh import plot_delaunay, plot_voronoi, plot_cell_indices
        try:
            voronoi_options = dict(centroid_style=None, color='rrrr') | voronoi_options
        except TypeError: # Python < 3.9
            voronoi_options, _options = dict(centroid_style=None, color='rrrr'), voronoi_options
            voronoi_options.update(_options)
        if isinstance(tessellation, Analysis):
            tessellation = tessellation.data
        if isinstance(tessellation, Partition):
            sampling = tessellation
            tessellation = sampling.tessellation
        else:
            sampling = None
        if isinstance(tessellation, TimeLattice):
            if tessellation.spatial_mesh is None:
                raise TypeError('no tessellation found; only time segments')
            tessellation = tessellation.spatial_mesh
        if isinstance(tessellation, Voronoi):
            if sampling is None:
                sampling = Partition(locations, tessellation)
        else:
            raise TypeError('tessellation type not supported: {}'.format(type(tessellation)))
        if delaunay_options:
            plot_delaunay(sampling, axes=axes, **delaunay_options)
        if voronoi_options:
            plot_voronoi(sampling, axes=axes, **voronoi_options)
        plot_cell_indices(tessellation, axes=axes, **kwargs)
        if title:
            axes.set_title(title)
        if aspect:
            axes.set_aspect(aspect)


__all__ = ['Mpl']




# -*- coding: utf-8 -*-

# Copyright © 2018-2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from ..core import *
from ..core.hdf5 import *
from ..core.analyses import abc
import os.path
import six
import pandas as pd
import pkg_resources
import platform
import time


class UseCaseWarning(UserWarning):
    pass
class IgnoredInputWarning(UserWarning):
    pass


class HelperBase(object):
    def __init__(self):
        self.analyses = {}
        self.metadata = {}

    def add_metadata(self, analysis, pkg_version=[]):
        if self.metadata is None:
            return
        analysis.metadata.update(self.metadata)
        if 'os' not in analysis.metadata:
            analysis.metadata['os'] = platform.system()
        if 'python' not in analysis.metadata:
            analysis.metadata['python'] = platform.python_version()
        for pkg in pkg_version:
            analysis.metadata['pkg'] = pkg_resources.get_distribution(pkg).version
        if 'tramway' not in analysis.metadata:
            analysis.metadata['tramway'] = pkg_resources.get_distribution('tramway').version
        if 'datetime' not in analysis.metadata:
            analysis.metadata['datetime'] = time.strftime('%Y-%m-%d %H:%M:%S UTC%z')


class Helper(object):

    def __init__(self):
        """
        """
        HelperBase.__init__(self)
        self.input_file = None
        self.analyses = None
        #self.plugins = None
        self.module = self.name = None
        self.setup = {}
        self.verbose = None # undefined
        self.input_label = self.output_label = None
        self._label_is_output = None
        self.comment = None

    def add_metadata(self, analysis, pkg_version=[]):
        if self.metadata is None:
            return
        HelperBase.add_metadata(self, analysis, pkg_version)
        if 'plugin' not in analysis.metadata:
            plugin = None
            try:
                if self.name in self.plugins:
                    plugin = self.name
            except AttributeError:
                pass
            if plugin:
                analysis.metadata['plugin'] = plugin

    def are_multiple_files(self, files):
        return isinstance(files, (tuple, list, frozenset, set))

    def prepare_data(self, data, labels=None, types=None, metadata=True, verbose=None, **kwargs):
        """
        Load the data if the input argument is a file (can be rwa or txt).
        If the input file is a trajectory file, trailing keyworded arguments are passed
        to :func:`load_xyt`.

        Isolate the requested artefacts if the input is an analysis tree or an rwa file
        and `labels` or `types` is defined.
        """
        if metadata is None or metadata is False:
            self.metadata = None
        if verbose is None:
            verbose = self.verbose
        if labels is None:
            labels = self.input_label
        if isinstance(data, pd.DataFrame):
            self.analyses = Analyses(data)
            if self.metadata:
                self.analyses.metadata.update(self.metadata)
                self.add_metadata(self.analyses)
                self.metadata = {}
        elif isinstance(data, abc.Analyses):
            self.analyses = data
        elif isinstance(data, six.string_types):
            if not os.path.isfile(data):
                raise OSError('file not found: {}'.format(data))
            self.input_file = [data]
            self.analyses = data = load_rwa(data, lazy=True, verbose=verbose)
        if isinstance(data, abc.Analyses):
            if not (labels is None and types is None):
                data = find_artefacts(data, types, labels)
                if self.input_label is None:
                    self.input_label = self.find(data[-1])
        return data

    def find(self, artefact, labels=None, return_subtree=False):
        if labels is None:
            labels = self.input_label
        found = False
        _labels = []
        _label = None
        analysis = self.analyses
        if analysis._data is artefact:
            found = True
        elif self.label_is_absolute(labels):
            for _label in labels:
                _labels.append(_label)
                analysis = analysis[_label]
                if analysis._data is artefact:
                    found = True
                    break
        else:
            assert labels is None
            available_labels = list(analysis.labels)
            # TODO: recursive search
            # for now, `labels` is required
            while available_labels and not available_labels[1:]:
                _label = available_labels[0]
                _labels.append(_label)
                analysis = analysis[_label]
                if analysis._data is artefact:
                    found = True
                    break
                available_labels = list(analysis.labels)
        if not found:
            #raise RuntimeError('artefact not found')
            _labels = []
            analysis = None
        if return_subtree:
            return _labels, analysis
        else:
            return _labels

    def output_file(self, output_file=None, suffix=None, extension=None):
        basename = None
        if output_file is None:
            if self.input_file:
                if self.input_file[1:]:
                    warnings.warn('formatting the output filename on basis of the first input filename', RuntimeWarning)
                if suffix is None and extension is None:
                    return self.input_file[0]
                else:
                    basename, = os.path.splitext(self.input_file[0])
        else:
            if suffix is None and extension is None:
                return output_file
            else:
                basename, = os.path.splitext(output_file)
        output_file = basename
        if basename:
            if suffix:
                if suffix[0].isalpha():
                    suffix = '_'+suffix
                output_file += suffix
            if extension:
                if extension[0] != '.':
                    extension = '.'+extension
                output_file += extension
        return output_file

    def label_is_absolute(self, label):
        return isinstance(label, (tuple, list))

    def labels(self, label=None, input_label=None, output_label=None, inplace=False, comment=None):
        # store comment
        self.comment = comment
        # define whether `label` is input or output label
        exclusive_labels_error = ValueError("multiple different values in exclusive arguments 'label', 'input_label' and 'output_label'")
        if label is not None:
            if self._label_is_output is True:
                if output_label is None:
                    output_label = label
                else:
                    raise ValueError("'label' and 'output_label' are both defined and are different")
            elif self._label_is_output is False:
                if input_label is None:
                    input_label = label
                else:
                    raise ValueError("'label' and 'input_label' are both defined and are different")
            else:
                if output_label is None:
                    if input_label is None:
                        # arbitrarily choose
                        input_label = label
                    else:
                        output_label = label
                else:
                    if input_label is None:
                        input_label = label
                    else:
                        if label == input_label or label == output_label:
                            # safely ignore `label`
                            pass
                        else:
                            raise exclusive_labels_error
            label = None
        # input labels necessarily form a path
        if input_label is None:
            self.explicit_input_label = False
        else:
            self.explicit_input_label = True
            if not self.label_is_absolute(input_label):
                input_label = [input_label]
        # if inplace, all the defined labels should equal
        if inplace:
            if output_label is None:
                label = input_label
            else:
                if self.label_is_absolute(output_label):
                    label = output_label
                    if not (input_label is None or output_label == input_label):
                        raise exclusive_labels_error
                else:
                    if input_label is None:
                        label = [output_label]
                    elif output_label == input_label[-1]:
                        label = input_label
                    else:
                        raise exclusive_labels_error
            output_label = input_label = label
        # return
        self.input_label = input_label
        self.output_label = output_label
        return input_label, output_label

    @property
    def inplace(self):
        return valid_label(self.input_label) \
            and self.label_is_absolute(self.output_label) \
            and self.input_label == self.output_label

    def plugin(self, name, plugins=None, verbose=None):
        if verbose is None:
            verbose = self.verbose
        if plugins is not None:
            self.plugins = plugins
        try:
            if self.plugins is None:    raise AttributeError
        except AttributeError:
            raise ValueError('no plugins defined')
        if verbose:
            self.plugins.verbose = True
        try:
            self.setup, self.module = self.plugins[name]
        except KeyError:
            raise KeyError('no such plugin: {}'.format(name))
        else:
            self.name = name
        return self.module, self.setup

    def insert_analysis(self, analysis_or_artefact, label=None, comment=None, anchor=None):
        if comment is None:
            comment = self.comment
        labels = None
        if isinstance(analysis_or_artefact, abc.Analyses):
            artefact = analysis_or_artefact.data
            analysis = analysis_or_artefact
        else:
            artefact = analysis_or_artefact
            analysis = Analyses(analysis_or_artefact)
        self.add_metadata(analysis)
        if self.analyses is None:
            raise RuntimeError('no root analysis')
        if anchor:
            labels, input_analysis = self.find(anchor, return_subtree=True,
                labels=label if self.label_is_absolute(label) else None)
            if input_analysis is None:
                raise ValueError('anchor not found')
            if label is None:
                label = self.output_label
            if self.label_is_absolute(label):
                if labels == label[:-1]:
                    label = label[-1]
                else:
                    raise ValueError('output labels do not match with anchor artefact')
        else:
            input_analysis = self.analyses
            if label is None:
                label = self.output_label
            if self.label_is_absolute(label):
                # output_label is absolute path; forget input_label
                labels = list(label)
                label = labels.pop() # terminal label
                for _label in labels:
                    input_analysis = input_analysis.instances[_label]
            else:
                labels = self.input_label
                if labels is not None:
                    # input_label is absolute path, even if it is not formatted correspondingly
                    if self.label_is_absolute(labels):
                        labels = list(labels)
                    else:
                        self.input_label = labels = [ labels ]
                    #
                    for _label in labels:
                        input_analysis = input_analysis.instances[_label]
        if label is not None and label in input_analysis.labels and self.inplace:
            # `input_analysis[label]` already exists
            if comment:
                input_analysis.comments[label] = comment
            input_analysis = input_analysis.instances[label]
            input_analysis.artefact = artefact
        else:
            label = input_analysis.autoindex(label)
            input_analysis.add(analysis, label=label, comment=comment)
        # self.analyses is implicitly updated
        if labels is None:
            return [label]
        else:
            labels.append(label)
            return labels

    def save_analyses(self, output_file=None, verbose=None, force=None, **kwargs):
        if verbose is None:
            verbose = self.verbose
        output_file = self.output_file(output_file)
        if output_file is None:
            return
        if force is None and bool(self.input_file):
            if self.are_multiple_files(self.input_file):
                input_files = list(self.input_file)
            else:
                input_files = [self.input_file]
            force = not input_files[1:] and input_files[0] == output_file
        save_rwa(output_file, self.analyses, verbose, force, **kwargs)



class AutosaveCapable(object):
    """
    deprecated
    """
    def __init__(self, rwa_file=None, autosave=True):
        self._autosave = autosave
        self._autosave_overwrite = None
        self.rwa_file = rwa_file
        self.save_options = dict(force=True)
        self._modified = None
        self._analysis_tree = None
        self._extra_artefacts = {} # deprecated
    @property
    def autosave(self):
        return self._autosave if self._autosave_overwrite is None else self._autosave_overwrite
    @property
    def save_on_completion(self):
        return self.autosave and (isinstance(self.autosave, bool) or self.autosave.endswith('completion'))
    @property
    def save_on_every_step(self):
        return self.autosave and isinstance(self.autosave, str) and self.autosave.endswith('every step')
    @property
    def force_save(self):
        return self.autosave and isinstance(self.autosave, str) and self.autosave.startswith('force')
    def save(self):
        if self._analysis_tree is None:
            raise RuntimeError("method 'save' called from outside the context")
        if self.rwa_file:
            save_rwa(self.rwa_file, self._analysis_tree, **self.save_options)
            if self._extra_artefacts: # deprecated
                #from rwa import HDF5Store
                f = HDF5Store(self.rwa_file, 'a')
                try:
                    for label, artefact in self._extra_artefacts.items():
                        f.poke(label, artefact)
                finally:
                    f.close()
            return True
    def autosaving(self, analysis_tree, on=None):
        if self.autosave:
            if on is not None:
                self._autosave_overwrite = on
            self._analysis_tree = analysis_tree
        return self
    def __enter__(self):
        self._modified = False
        return self
    def __exit__(self, exc_type, exc_value, exc_traceback):
        self._autosave_overwrite = None
        if self._modified:
            if exc_type is None:
                if self.save_on_completion:
                    self.save()
                    # unload
                    self._analysis_tree = None
            else:
                if self.force_save:
                    self.save()
                    self._analysis_tree = None # useless as an exception will be raised anyway
        # reset
        self._modified = None
    @property
    def modified(self):
        if self._modified is None:
            raise RuntimeError("property 'modified' called from outside the context")
        return self._modified
    @modified.setter
    def modified(self, b):
        if self._modified is None:
            raise RuntimeError("property 'modified' called from outside the context")
        if b is not True:
            raise ValueError("property 'modified' can only be set to True")
        self._modified = b




# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: Jean-Baptiste Masson

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


# Import Libraries and model
## files management
import argparse
from   os.path import abspath
import gc
import os

## general
import numpy as np
import pandas as pd

## image
try:
    from skimage import io
except ImportError:
    raise ImportError('please install scikit-image>=0.14.2')
from os.path import split
from skimage.feature import peak_local_max
from skimage import data, img_as_float

## keras stuff
try: # recent Keras
	from tensorflow.keras.preprocessing.image import ImageDataGenerator
	from tensorflow.keras.callbacks import ModelCheckpoint, ReduceLROnPlateau, Callback
	#from tensorflow.keras.models import Model
	from .tf import Model
	from tensorflow.keras.layers import Input, Activation, UpSampling2D, Convolution2D, Convolution2DTranspose, MaxPooling2D, BatchNormalization, Dropout, Lambda, concatenate
	from tensorflow.keras import optimizers, losses
	from tensorflow.keras.utils import multi_gpu_model
	from tensorflow.keras.backend import clear_session
except ImportError:
	from keras.preprocessing.image import ImageDataGenerator
	from keras.callbacks import ModelCheckpoint, ReduceLROnPlateau, Callback
	from keras.models import Model
	from keras.layers import Input, Activation, UpSampling2D, Convolution2D, Convolution2DTranspose, MaxPooling2D, BatchNormalization, concatenate
	from keras import optimizers, losses
	from keras.utils import multi_gpu_model
	from keras.layers.core import Dropout, Lambda
	from keras.backend import clear_session
## local stuff
from .utility_function_inference import *

#########################################################################################
#########################################################################################
#########################################################################################
#########################################################################################
def get_parameters_mean_std(mean_std_file):

	file       = open(mean_std_file, "r")
	mean_image = np.float32(file.readline())
	std_image  = np.float32(file.readline())
	file.close()

	return mean_image,std_image
#########################################################################################
#########################################################################################
#########################################################################################
#########################################################################################
def preprocess_prediction(high_res_prediction,threshold):

	high_res_prediction[high_res_prediction<threshold] = 0
	image_high_res      = high_res_prediction.astype('uint16')

	image_high_res      = np.squeeze(image_high_res)
	high_res_prediction = np.squeeze(high_res_prediction)
	
	return high_res_prediction, image_high_res 


#########################################################################################
#########################################################################################
#########################################################################################
#########################################################################################
def remove_anomalies(test):

	test_diff = np.diff(test,axis=0)
	test_diff = np.linalg.norm(test_diff, axis =1)
	II        = np.where(test_diff<=1.5)
	output    = np.delete(test, II, axis =0)

	return output

#########################################################################################
#########################################################################################
#########################################################################################
#########################################################################################
def dummy_get_rough_localisation_one_image(image_high_res, min_distance_peak,threshold_abs):

	(M,N) = image_high_res.shape
	liste   = []

	test = peak_local_max( image_high_res,threshold_abs=threshold_abs, min_distance=min_distance_peak)
	temp = test[:,0]
	test[:,[0, 1]] = test[:,[1, 0]]
	test = remove_anomalies(test)
	liste.append(test)
	del test, temp
	#del test
	return liste
#########################################################################################
#########################################################################################
#########################################################################################
#########################################################################################

def get_position_from_predicted_one_image(high_res_prediction, liste, marge, k):

	(M,N)        = high_res_prediction.shape
	xxx          = np.arange(M)
	xxx          = np.expand_dims(xxx,axis=1)
#	yyy          = xxx.transpose()

	yyy          = np.arange(N)
	yyy          = yyy.transpose()
	yyy          = np.expand_dims(yyy,axis=0)


	high_x       = np.multiply(xxx[np.newaxis,:],high_res_prediction[:,:])
	high_y       = np.multiply(yyy[np.newaxis,:],high_res_prediction[:,:])

	liste_output = []

#	for k in range(K):
		#print(k)
		#print("%i\t" % k)
	liste_loc      = liste[0]
	(KK, LL)       = liste_loc.shape
	point_high_res = np.zeros((KK,LL+1))
	if KK<=0:
		raise ValueError('no iteration; magnification might be too low')
	for kk in range(KK):
		ii        = liste_loc[kk,0]
		jj        = liste_loc[kk,1]

		min_x_loc = np.maximum(jj-marge,0)
		max_x_loc = np.minimum(jj+marge+1,M)
		min_y_loc = np.maximum( ii-marge, 0)
		max_y_loc = np.minimum(ii+marge+1, N)

		proba     = high_res_prediction[min_x_loc:max_x_loc, min_y_loc:max_y_loc]
		proba_x   = high_x[min_x_loc:max_x_loc, min_y_loc:max_y_loc]
		proba_y   = high_y[min_x_loc:max_x_loc, min_y_loc:max_y_loc]	

		#proba     = high_res_prediction[jj-marge:jj+marge+1, ii-marge:ii+marge+1]
		#proba_x   = high_x[jj-marge:jj+marge+1, ii-marge:ii+marge+1]
		sum_proba = np.sum(proba,axis=None)
			#print((k,kk,sum_proba,np.sum(proba_x,axis=None),np.sum(proba_y,axis=None) ))
			
		point_high_res[kk,0] = k
		point_high_res[kk,1] = np.sum(proba_x,axis=None)/sum_proba
		point_high_res[kk,2] = np.sum(proba_y,axis=None)/sum_proba

			#print((liste_loc[kk,0],point_high_res[kk,1],liste_loc[kk,1],point_high_res[kk,2]))
	point_high_res[:,[1,2]] = point_high_res[:,[2, 1]]
	liste_output.append(point_high_res)
	del proba, proba_x, proba_y, sum_proba, liste_loc, point_high_res
	#print(K)
	#if K==1:
	#	point_high_res[:,0] = nb_out

	position         = pd.DataFrame(np.concatenate(liste_output))
	position.dropna(how='all')
	position.columns = ['nb','x','y']
	position['nb']   = position['nb'].astype(int)
	
	return position
#########################################################################################
#########################################################################################
#########################################################################################
#########################################################################################

def print_position_files(position,path,root_name, bool_header):

	#file = open(path + '/' + 'position_' root_name + '.txt','w+' )
	full_path= path + '/' + 'position_' + root_name + '.txt'
	if bool_header:
		position.to_csv(full_path,index=None, sep=',') 
	else:
		position.to_csv(full_path,index=None,header=None, sep=',') 

#########################################################################################
#########################################################################################
#########################################################################################
#########################################################################################
def save_trimmed_original_image_magnified_for_testing_purposes(stack,magnification,n):

	path_split           = os.path.split(stack)
	name_split           = os.path.splitext(path_split[1])

	Images = io.imread(stack)
	Images = Images[:,n:-n,n:-n]
	# get dataset dimensions
	(K, M, N) = Images.shape
	high_res = np.zeros((K,M*magnification, N*magnification))
	for i in range(K):
		high_res[i,:,:] = np.kron(Images[i,:,:], np.ones((magnification,magnification)))  
	(K, M, N) = high_res.shape
	image_high_res = high_res.astype('uint16')

	io.imsave(path_split[0] + '/' + name_split[0] + '_magnified.tiff', image_high_res)


	return 1

#########################################################################################
#########################################################################################
#########################################################################################
#########################################################################################


def pre_process_one_image(Image, index_split, M_extend, N_extend, M_original, N_original, magnification,mean_image,std_image):

	#(M,N)    = Images.shape
	Image_extend      = np.zeros((M_extend,N_extend))
	Image_extend[0:M_original,0:N_original] = Image
	liste_Image_extend_magnificaton = []
	for x in index_split:
		high_res = np.zeros((M_extend*magnification, N_extend*magnification))
		high_res = np.kron(  Image_extend[x[0]:x[1],x[2]:x[3]], np.ones((magnification,magnification)))
		high_res = normalize_one_image(high_res)
		high_res = normalize_stack_whitening_from_trained_data(high_res, mean_image, std_image)
		liste_Image_extend_magnificaton.append(high_res)

	Image_to_stack = np.asarray(liste_Image_extend_magnificaton)
	(K_stack, _,_) = Image_to_stack.shape	


	return Image_to_stack, K_stack

#########################################################################################
#########################################################################################
#########################################################################################
#########################################################################################

def reassemble_one_image(Image_to_stack,K_stack, empty_Image_high_res, index_reconstitute_mag, M_cut, N_cut, n_high):

	liste_loc = np.split(Image_to_stack,K_stack)

	for indice,x in enumerate(index_reconstitute_mag,0):
		#print(x[0],x[1],x[2],x[3])
		empty_Image_high_res[x[0]:x[1],x[2]:x[3]] = np.squeeze(liste_loc[indice][0,n_high:-n_high,n_high:-n_high])
	empty_Image_high_res                      = empty_Image_high_res[0:M_cut,0:N_cut]


	return empty_Image_high_res

#########################################################################################
#########################################################################################
#########################################################################################
#########################################################################################


def define_all_indexes_for_image_slicing(Images_example, M_theo, N_theo, n, magnification):

	##################################################################
	(M,N)             = Images_example.shape
	II_max            = int(np.ceil((M-M_theo)/(M_theo-2*n)))+1
	JJ_max            = int(np.ceil((N-N_theo)/(N_theo-2*n)))+1
	M_extend          = (II_max ) * M_theo - n
	N_extend          = (JJ_max ) * N_theo - n
	M_extend_high_res = M_extend*magnification 
	N_extend_high_res = N_extend*magnification 
	##################################################################
	I_start           = np.arange(0,II_max)*(M_theo-2*n)
	I_end             = I_start + M_theo
	J_start           = np.arange(0,JJ_max)*(N_theo-2*n) 
	J_end             = J_start + N_theo
	################################################################
	I_start_rec       = np.arange(0,II_max)*(M_theo-2*n)
	I_end_rec         = I_start_rec + (M_theo-2*n)
	J_start_rec       = np.arange(0,JJ_max)*(N_theo-2*n)
	J_end_rec         = J_start_rec + (N_theo-2*n)
	###############################################################
	I_start_rec_mag   = np.arange(0,II_max)*(M_theo-2*n)*magnification
	I_end_rec_mag     = I_start_rec_mag + (M_theo-2*n)*magnification
	J_start_rec_mag   = np.arange(0,JJ_max)*(N_theo-2*n)*magnification
	J_end_rec_mag     = J_start_rec_mag + (N_theo-2*n)*magnification
	##################################################################
	index_split       = []
	index_reconstitute = []
	index_reconstitute_mag = []
	
	for i in range(II_max):
		for j in range(JJ_max):
			index_split.append([I_start[i], I_end[i], J_start[j],J_end[j]])
			index_reconstitute.append([I_start_rec[i], I_end_rec[i], J_start_rec[j],J_end_rec[j]])
			index_reconstitute_mag.append([I_start_rec_mag[i], I_end_rec_mag[i], J_start_rec_mag[j],J_end_rec_mag[j]])
	##################################################################
	M_cut_high_res = M*magnification - 2*n*magnification
	N_cut_high_res = N*magnification - 2*n*magnification
	n_high         = n*magnification
	##################################################################
	


	return M_extend, N_extend,M_extend_high_res, N_extend_high_res, index_split, index_reconstitute, index_reconstitute_mag, M_cut_high_res, N_cut_high_res,n_high 
#########################################################################################
#########################################################################################
#########################################################################################
#########################################################################################
def Inference(stack,magnification,  weights, mean_image, std_image, M_theo, N_theo,n, threshold, min_distance_peak, threshold_abs, marge, bool_GPU ):
	## name of the stack with full path
	## factor of magnification
	## full path to wieghts file
	## mean of the training images
	## std of the training images

	Images = io.imread(stack)
	(K_original,M_original,N_original) = Images.shape
	if bool_GPU:
		model = generate_network_keras_multi_GPUs((M_theo*magnification, N_theo*magnification, 1))
	else:
		model = generate_network_keras_for_inference_one_GPU((M_theo*magnification, N_theo*magnification, 1))

	model.load_weights(weights)

	## all usefull indexes
	M_extend, N_extend,M_extend_high_res, N_extend_high_res, index_split,\
	index_reconstitute, index_reconstitute_mag, M_cut_high_res, N_cut_high_res, \
	n_high = define_all_indexes_for_image_slicing(np.squeeze(Images[0,:,:]), M_theo, N_theo, n, magnification)
	## empty recipient
	empty_Image_high_res = np.zeros((M_extend_high_res - 2*n*magnification , N_extend_high_res - 2*n*magnification))

	##
	liste_position = []
	#high_res_prediction = np.zeros((K_original, M_cut_high_res, N_cut_high_res))
	#liste_high_res = []
	for i in range(K_original):
		if i%100==0:
			print(i)
		Image          		                = np.squeeze(Images[i,:,:])
		Image_to_stack,K_stack              = pre_process_one_image(Image, index_split, M_extend, N_extend, M_original, N_original, magnification,mean_image,std_image)
		Image_to_stack                      = np.squeeze(model.predict(np.expand_dims(Image_to_stack,axis=3), batch_size=1))
		high_res                            = reassemble_one_image(Image_to_stack,K_stack, empty_Image_high_res, index_reconstitute_mag, M_cut_high_res, N_cut_high_res, n_high)
		high_res, image_high_res            = preprocess_prediction(high_res,threshold)
		liste_low_res                       = dummy_get_rough_localisation_one_image(image_high_res, min_distance_peak,threshold_abs)
		position                            = get_position_from_predicted_one_image(high_res, liste_low_res, marge,i)
		#position                            = get_position_from_predicted(np.expand_dims(high_res,axis=0), liste_low_res, marge,i)
		liste_position.append(position)
		#high_res_prediction[i,:,:] = high_res
		#liste_high_res.append(high_res)
		del Image, Image_to_stack, K_stack, high_res, image_high_res,liste_low_res, position

	#high_res_presdiction = np.stack(liste_high_res)
	position = pd.concat(liste_position)
	#del high_res_norm
	high_res_presdiction = None

	return high_res_prediction, position
#########################################################################################
#########################################################################################
#########################################################################################
#########################################################################################
def generate_network_keras_for_inference_one_GPU(input_dim):

    input_         = Input (shape = (input_dim))
    full_CNN       = CNN(input_,'CNN')
    model          = Model (inputs= input_, outputs=full_CNN)
    opt            = optimizers.Adam(lr= 0.001)
    model.compile(optimizer=opt, loss = Simplest_loss(input_dim))
  

    return model


#########################################################################################
#########################################################################################
#########################################################################################
#########################################################################################
def generate_network_keras_multi_GPUs(input_dim):
    
    input_         = Input (shape = (input_dim))
    full_CNN       = CNN(input_,'CNN')
    model          = Model (inputs= input_, outputs=full_CNN)
    parallel_model = multi_gpu_model(model,gpus=4)
    opt            = optimizers.Adam(lr= 0.001)
 
    parallel_model.compile(optimizer=opt, loss = Simplest_loss(input_dim))
    #model.compile(optimizer=opt, loss = Simplest_loss(input_dim))

    return parallel_model

#########################################################################################
#########################################################################################
#########################################################################################
#########################################################################################


def CNN(input,names):

    # start 
    Features1  = module_CNN(16,5,5,names+'F1',0)(input)
    Features2  = module_CNN(16,5,5,names+'F2',0)(Features1)
    # reduce size convole
    Features3  = module_CNN_to_down(32,5,5,names+'F3',0)(Features2)
    Features4  = module_CNN_to_down(64,5,5,names+'F4',0)(Features3)
    Features5  = module_CNN_to_down(128,5,5,names+'F5',0)(Features4)
    Features6  = module_CNN_to_down(256,5,5,names+'F6',0)(Features5)
    # convolve
    Features7  = module_CNN(512,5,5,names+'F7',1)(Features6)
    # increase size concatenae covolve
    Features8  = module_CNN_to_up(256,5,5,names+'F8',1)(Features7)
    Features9  = concatenate([Features8, Features5])
   
    Features10  = module_CNN_to_up(128,5,5,names+'F9',1)(Features9)
    Features11  = concatenate([Features10, Features4])

    Features12  = module_CNN_to_up(64,5,5,names+'F10',1)(Features11)
    Features13 = concatenate([Features12, Features3])

    Features14  = module_CNN_to_up(32,5,5,names+'F11',1)(Features13)
    Features15 = concatenate([Features14, Features2])
    #convolve    
    Features16 = module_CNN(16,5,5,names+'F12',0)(Features15)
    Features17 = module_CNN(16,5,5,names+'F13',0)(Features16)
    #get_densities
    full_CNN   = Convolution2D(1, kernel_size=(1, 1), strides=(1, 1), padding="same",\
                                  activation="linear", use_bias = False,\
                                  kernel_initializer="Orthogonal",name='Prediction')(Features17)

    return full_CNN

#########################################################################################
#########################################################################################
#########################################################################################
def module_CNN_to_down(nb_filter, rk, ck, name, boolean_dropout):
    def f(input):
        layer                = Convolution2D(nb_filter, kernel_size=(rk, ck), strides=(1,1),\
                               padding="same", use_bias=False,\
                               kernel_initializer="Orthogonal",name='conv-'+name)(input)
        layer_norm           = BatchNormalization(name='BN-'+name)(layer)
        layer_norm_relu      = Activation(activation = "relu",name='Relu-'+name)(layer_norm)
        if boolean_dropout:
            layer_norm_relu =  Dropout(0.1) (layer_norm_relu)
        layer_norm_relu_pool = MaxPooling2D(pool_size=(2,2),name=name+'Pool1')(layer_norm_relu)


        return layer_norm_relu_pool
    return f

#########################################################################################
#########################################################################################
#########################################################################################

def module_CNN(nb_filter, rk, ck, name, boolean_dropout):
    def f(input):
        layer           = Convolution2D(nb_filter, kernel_size=(rk, ck), strides=(1,1),\
                               padding="same", use_bias=False,\
                               kernel_initializer="Orthogonal",name='conv-'+name)(input)
        layer_norm      = BatchNormalization(name='BN-'+name)(layer)
        layer_norm_relu = Activation(activation = "relu",name='Relu-'+name)(layer_norm)
        if boolean_dropout:
            layer_norm_relu =  Dropout(0.1) (layer_norm_relu)

        return layer_norm_relu
    return f
#########################################################################################
#########################################################################################
#########################################################################################
def module_CNN_to_up(nb_filter, rk, ck, name, boolean_dropout):
    def f(input):
        layer           = Convolution2DTranspose(nb_filter, kernel_size=(rk, ck), strides=(2,2),\
                               padding="same", use_bias=False,\
                               kernel_initializer="Orthogonal",name='conv-'+name)(input)
        layer_norm      = BatchNormalization(name='BN-'+name)(layer)
        layer_norm_relu = Activation(activation = "relu",name='Relu-'+name)(layer_norm)
        if boolean_dropout:
            layer_norm_relu =  Dropout(0.1) (layer_norm_relu)

        return layer_norm_relu
    return f
#########################################################################################
#########################################################################################
#########################################################################################
#########################################################################################

def Simplest_loss(input_shape):
    def MSE_densities(ground_truth, prediction):

        loss_loc = 10*losses.mean_squared_error(ground_truth,prediction)
        
        return loss_loc
    return MSE_densities

#########################################################################################
#########################################################################################
#########################################################################################
#########################################################################################
#########################################################################################


if __name__ == '__main__':

	## parameters hard coded
	magnification     = 10
	threshold         = 0
	min_distance_peak = 2
	marge             = 3
	bool_header       = 1
	threshold_abs     = 1.
	M_theo            = 64
	N_theo            = M_theo
	n                 = 2 

	## clean evrything
	##
	gc.collect()
	clear_session()
	##
	## get the prameters from command line
	parser = argparse.ArgumentParser()
	parser.add_argument('--stack',    help="path to tiff stack")
	parser.add_argument('--weights',  help="name of the weights file")
	parser.add_argument('--nb_GPU',   type=int, help="boolean 0= 1 GPU , 1 = multiple GPU")
	parser.add_argument('--mean_std', help="path to the mean_std.txt file")
	args   = parser.parse_args()
	##
	##

	path_split              = os.path.split(abspath(args.stack))
	name_split              = os.path.splitext(path_split[1])
	path_weights            = abspath(args.weights)
	mean_std_file           = abspath(args.mean_std)
	bool_GPU                = bool(args.nb_GPU)
	stack                   = abspath(args.stack)

    ## at the moment it is in separate files it could be stored depending on the use 
	mean_image,std_image    = get_parameters_mean_std(mean_std_file)

	## to be commented when applied in tramway 
	save_trimmed_original_image_magnified_for_testing_purposes(stack,magnification,n)

	## when in tramway it is not necesary to output high_res_prediction , is might be usefull anyway so to be put in option
	high_res_prediction, position = Inference(stack,magnification,  path_weights, mean_image, std_image, M_theo, N_theo,n, threshold, min_distance_peak, threshold_abs, marge,bool_GPU )
	
	## to be extended to rw files
	print_position_files(position,path_split[0],name_split[0], bool_header)

	## just for talks output high resolution version of the raw data with the distribution of position
	#image                   = high_res_prediction.astype('uint16')
	#io.imsave(path_split[0] + '/' + 'predicted.tiff', image)



#########################################################################################
#########################################################################################
#########################################################################################
#########################################################################################



# -*- coding: utf-8 -*-

# Copyright © 2017-2019, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from .base import *
from warnings import warn
from math import pi, log
import numpy as np
import pandas as pd
from scipy.optimize import minimize
from collections import OrderedDict


setup = {'name':    ('unsmooth.d', 'degraded.d'),
        'provides': 'd',
        'arguments': OrderedDict((
        ('localization_error',  ('-e', dict(type=float, help='localization precision (see also sigma; default is 0.03)'))),
        ('jeffreys_prior',      ('-j', dict(action='store_true', help="Jeffreys' prior"))),
        ('min_diffusivity',     dict(type=float, help='minimum diffusivity value allowed')))),
        'cell_sampling':    'individual'}


def d_neg_posterior(diffusivity, cell, sigma2, jeffreys_prior, dt_mean, \
    min_diffusivity):
    """
    Adapted from InferenceMAP's *dPosterior* procedure:

    .. code-block:: c++

        for (int i = 0; i < ZONES[CURRENT_ZONE].translocations; i++) {
            const double dt = ZONES[CURRENT_ZONE].dt[i];
            const double dx = ZONES[CURRENT_ZONE].dx[i];
            const double dy = ZONES[CURRENT_ZONE].dy[i];

            D_bruit = LOCALIZATION_ERROR*LOCALIZATION_ERROR/dt;

            result += - log(4.0*PI*(D[0]+D_bruit)*dt) - (dx*dx)/(4.0*(D[0]+D_bruit)*dt) - (dy*dy)/(4.0*(D[0]+D_bruit)*dt);
        }

        if (JEFFREYS_PRIOR == 1) {
            result += - (D[0]*ZONES[CURRENT_ZONE].dtMean + LOCALIZATION_ERROR*LOCALIZATION_ERROR);
        }
        return -result;

    """
    if diffusivity < min_diffusivity:# and not np.isclose(diffusivity, min_diffusivity):
        warn(DiffusivityWarning(diffusivity, min_diffusivity))
    noise_dt = sigma2
    if cell.cache is None:
        cell.cache = np.sum(cell.dr * cell.dr, axis=1) # dx**2 + dy**2 + ..
    n = len(cell) # number of translocations
    D_dt = 4. * (diffusivity * cell.dt + noise_dt) # 4*(D+Dnoise)*dt
    if np.any(D_dt <= 0):# or np.any(np.isclose(D_dt, 0)):
        raise RuntimeError('negative diffusion')
    d_neg_posterior = n * log(pi) + np.sum(np.log(D_dt)) # sum(log(4*pi*Dtot*dt))
    d_neg_posterior += np.sum(cell.cache / D_dt) # sum((dx**2+dy**2+..)/(4*Dtot*dt))
    if jeffreys_prior:
        d_neg_posterior += 2. * log(diffusivity * dt_mean + sigma2)
    return d_neg_posterior


def infer_D(cells, localization_error=None, jeffreys_prior=False, min_diffusivity=None, **kwargs):
    """
    Infer the local diffusivity in each cell.

    This function is similar to `infer_smooth_D` without the smoothing prior on
    the diffusivity.
    The problem could be solved analytically, instead of minimizing a cost
    function [TODO].
    """
    if isinstance(cells, Distributed): # multiple cells
        localization_error = cells.get_localization_error(kwargs, 0.03, True, \
                localization_error=localization_error)
        args = (localization_error, jeffreys_prior, min_diffusivity)
        inferred = { i: infer_D(c, *args, **kwargs) for i, c in cells.items() }
        inferred = pd.DataFrame({'diffusivity': pd.Series(inferred)})
        return inferred
    else: # single cell
        cell = cells
        # sanity checks
        if not bool(cell):
            raise ValueError('empty cell')
        if cell.dr.shape[1] == 0:
            raise ValueError('translocation array has no column')
        if cell.dt.shape[1:]:
            raise ValueError('time deltas are structured in multiple dimensions')
        # ensure that translocations are properly oriented in time
        if not np.all(0 < cell.dt):
            warn('translocation dts are not all positive', RuntimeWarning)
            cell.dr[cell.dt < 0] *= -1.
            cell.dt[cell.dt < 0] *= -1.
        #assert not np.isclose(np.mean(cell.dt), 0)
        # initialize the diffusivity value and cell cache
        dt_mean = np.mean(cell.dt)
        D_initial = np.mean(cell.dr * cell.dr) / (2. * dt_mean)
        cell.cache = None # clear the cache (optional, since `run` also clears it)
        # parametrize the optimization procedure
        if min_diffusivity is not False:
            if min_diffusivity is None:
                noise_dt = localization_error
                min_diffusivity = (1e-16-noise_dt) / np.max(cell.dt)
            kwargs['bounds'] = [(min_diffusivity,None)]
        # run the optimization
        result = minimize(d_neg_posterior, D_initial, \
            args=(cell, localization_error, jeffreys_prior, dt_mean, min_diffusivity), \
            **kwargs)
        # return the resulting optimal diffusivity value
        return result.x[0]




# -*- coding: utf-8 -*-

# Copyright © 2020-2021, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


import bokeh.plotting as plt
from bokeh.models.ranges import Range1d
from bokeh.models import ColorBar, LinearColorMapper, BasicTicker
from bokeh.layouts import row
from tramway.core import *
from tramway.tessellation.base import Partition, format_cell_index, Tessellation, Delaunay
from tramway.inference.base import FiniteElements
from tramway.plot.mesh import _graph_theme
import tramway.plot.map as mplt
import matplotlib as mpl
from math import tan, atan2, degrees, radians
import numpy as np
import pandas as pd
import itertools
from warnings import warn
import scipy.sparse as sparse
from collections import defaultdict


_long_colour_names = {
        'r':    'red',
        'g':    'green',
        'b':    'blue',
        'c':    'cyan',
        'm':    'magenta',
        'y':    'yellow',
        'k':    'black',
        }
def long_colour_name(c):
    if isinstance(c, str):
        c = _long_colour_names.get(c, c)
    return c

def plot_points(cells, style='circle', size=2, color=None, figure=None, **kwargs):
    """
    Plot molecule locations.

    If no colour is explicitly defined, colouring is based on the cell index.

    Arguments:

        cells (Partition):
            full partition

        style (str):
            point marker style

        size (float):
            point marker size (in screen units)

        color (str or numpy.ndarray):
            cell colours

        figure (bokeh.plotting.figure.Figure):
            figure handle

    returns:

        list: handles of the various clouds of points

    """
    # for compatibility with tramway.plot.mesh.plot_points
    try:
        kwargs.pop('min_count')
    except KeyError:
        pass

    if isinstance(cells, np.ndarray):
        points = cells
        label = None
    elif isinstance(cells, pd.DataFrame):
        points = cells[['x','y']].values
        label = None
    elif isinstance(cells, Partition):
        points = cells.descriptors(cells.points, asarray=True)
        label = cells.cell_index
        npts = points.shape[0]
        ncells = cells.location_count.size
        # if label is not a single index vector, convert it following
        # tessellation.base.Delaunay.cell_index with `preferred`='force index'.
        merge = mplt.nearest_cell(points, cells.tessellation.cell_centers)
        label = format_cell_index(cells.cell_index, format='array', \
            select=merge, shape=(npts, ncells))

    if isstructured(points):
        x = points['x']
        y = points['y']
    else:
        x = points[:,0]
        y = points[:,1]

    if figure is None:
        assert False
        try:
            figure = plt.curplot() # does not work
        except:
            figure = plt.figure()

    handles = []

    if label is None:
        if color is None:
            color = 'k'
        elif isinstance(color, (pd.Series, pd.DataFrame)):
            raise NotImplementedError
            color = np.asarray(color)
        if isinstance(color, np.ndarray):
            raise NotImplementedError
            cmin, cmax = np.min(color), np.max(color)
            color = (color - cmin) / (cmax - cmin)
            cmap = plt.get_cmap()
            color = [ cmap(c) for c in color ]
        else:
            color = long_colour_name(color)
        h = figure.scatter(x, y, marker=style, color=color, size=size, **kwargs)
        handles.append(h)
    else:
        L = np.unique(label)
        if color in [None, 'light']:
            if color == 'light' and 'alpha' not in kwargs:
                kwargs['alpha'] = .2
            if 2 < len(L):
                color = __colors__
                color = ['gray'] + \
                    list(itertools.islice(itertools.cycle(color), len(L)))
            elif len(L) == 2:
                color = ['gray', 'k']
            else:   color = 'k'
        elif isinstance(color, str) and L.size==1:
            color = [color]
        for i, l in enumerate(L):
            clr_i = long_colour_name(color[i])
            h = figure.scatter(x[label == l], y[label == l],
                    marker=style, color=clr_i, size=size, **kwargs)
            handles.append(h)

    ## resize window
    #try:
    #    figure.x_range = Range1d(*cells.bounding_box['x'].values)
    #    figure.y_range = Range1d(*cells.bounding_box['y'].values)
    #except AttributeError:
    #    pass
    #except ValueError:
    #    traceback.print_exc()

    return handles

def plot_trajectories(trajs, color=None, loc_style='circle', figure=None, **kwargs):
    """
    Plot trajectories.

    If no colour is explicitly defined, colouring is based on the trajectory index.

    Arguments:

        trajs (pandas.DataFrame):
            full partition

        loc_style (str):
            location marker style

        loc_size (float):
            location marker size (in screen units)

        color (str or numpy.ndarray):
            trajectory colours

        figure (bokeh.plotting.figure.Figure):
            figure handle

    returns:

        list: handles of the glyphs

    """
    loc_kwargs = {}
    line_kwargs = {}
    for attr in dict(kwargs):
        if attr.startswith('loc_'):
            loc_kwargs[attr[4:]] = kwargs.pop(attr)
        if attr.startswith('line_'):
            if attr == 'line_width':
                # compatibility for old bokeh versions
                line_kwargs[attr] = kwargs.pop(attr)
                continue
            line_kwargs[attr[5:]] = kwargs.pop(attr)

    loc_kwargs.update(kwargs)
    line_kwargs.update(kwargs)

    lines_x, lines_y = [], []
    for _, df in trajs.groupby([trajs['n']]):
        lines_x.append(df['x'].values)
        lines_y.append(df['y'].values)

    if figure is None:
        assert False
        try:
            figure = plt.curplot() # does not work
        except:
            figure = plt.figure()

    handles = []

    if color is None:
        ntrajs = len(lines_x)
        if color in [None, 'light']:
            if color == 'light' and 'alpha' not in kwargs:
                kwargs['alpha'] = .2
            if 2 < ntrajs:
                color = __colors__
                color = ['gray'] + \
                    list(itertools.islice(itertools.cycle(color), ntrajs))
            elif ntrajs == 2:
                color = ['gray', 'k']
            else:   color = 'k'
        elif isinstance(color, str) and ntrajs==1:
            color = [color]
        for i, line in enumerate(zip(lines_x, lines_y)):
            line_x, line_y = line
            clr_i = long_colour_name(color[i])
            h = figure.line(line_x, line_y, color=clr_i, **line_kwargs)
            handles.append(h)
            h = figure.scatter(line_x, line_y, color=clr_i, marker=loc_style, **loc_kwargs)
            handles.append(h)
    else:
        color = long_colour_name(color)
        h = figure.multi_line(lines_x, lines_y, color=color, **line_kwargs)
        handles.append(h)
        h = figure.scatter(list(itertools.chain(*lines_x)), list(itertools.chain(*lines_y)),
                color=color, marker=loc_style, **loc_kwargs)
        handles.append(h)

    return handles



def scalar_map_2d(cells, values, clim=None, figure=None, delaunay=False,
        colorbar=True, colormap=None, unit=None, clabel=None, colorbar_figure=None,
        xlim=None, ylim=None, try_fix_corners=True, **kwargs):
    """
    Plot an interactive 2D scalar map as a colourful image.

    Arguments:

        cells (Partition): spatial description of the cells

        values (pandas.Series or pandas.DataFrame): feature value at each cell/bin,
            encoded into a colour

        clim (2-element sequence): passed to :func:`~matplotlib.cm.ScalarMappable.set_clim`

        figure (bokeh.plotting.figure.Figure): figure handle

        delaunay (bool or dict): overlay the Delaunay graph; if ``dict``, options are passed
            to :func:`~tramway.plot.bokeh.plot_delaunay`

        colorbar (bool or str or dict): add a colour bar; if ``dict``, options are passed to
            :func:`~bokeh.models.ColorBar`

        unit/clabel (str): colorbar label, usually the unit of displayed feature

        colormap (str): colormap name; see also https://matplotlib.org/users/colormaps.html

        xlim (2-element sequence): lower and upper x-axis bounds

        ylim (2-element sequence): lower and upper y-axis bounds

    """
    if isinstance(values, pd.DataFrame):
        feature_name = values.columns[0]
        values = values.iloc[:,0] # to Series
    else:
        feature_name = None

    if figure is None:
        assert False
        figure = plt.figure()
    glyph_renderers = []

    ids = []
    polygons = []

    xy = cells.tessellation.cell_centers
    if not xlim or not ylim:
        xy_min, _, xy_max, _ = mplt._bounding_box(cells, xy)
        if not xlim:
            xlim = (xy_min[0], xy_max[0])
        if not ylim:
            ylim = (xy_min[1], xy_max[1])
    ix = np.arange(xy.shape[0])
    try:
        vertices, cell_vertices, Av = mplt.box_voronoi_2d(cells.tessellation, xlim, ylim)
    except AssertionError:
        warn('could not fix the borders', RuntimeWarning)
        vertices, cell_vertices, Av = cells.tessellation.vertices, cells.tessellation.cell_vertices, cells.tessellation.vertex_adjacency.tocsr()
    try:
        ok = 0 < cells.location_count
    except (KeyboardInterrupt, SystemExit):
        raise
    except:
        print(traceback.format_exc())
        ok = np.ones(ix.size, dtype=bool)
    if cells.tessellation.cell_label is not None:
        ok = np.logical_and(ok, 0 < cells.tessellation.cell_label)
    map_defined = np.zeros_like(ok)
    map_defined[values.index] = True
    ok[np.logical_not(map_defined)] = False
    ok[ok] = np.logical_not(np.isnan(values.loc[ix[ok]].values))
    for i in ix[ok]:
        vs = cell_vertices[i].tolist()
        # order the vertices so that they draw a polygon
        v0 = v = vs[0]
        vs = set(vs)
        _vertices = []
        #vvs = [] # debug
        while True:
            _vertices.append(vertices[v])
            #vvs.append(v)
            vs.remove(v)
            if not vs:
                break
            ws = set(Av.indices[Av.indptr[v]:Av.indptr[v+1]]) & vs
            if not ws:
                ws = set(Av.indices[Av.indptr[v0]:Av.indptr[v0+1]]) & vs
                if ws:
                    _vertices = _vertices[::-1]
                else:
                    #print((v, vs, vvs, [Av.indices[Av.indptr[v]:Av.indptr[v+1]] for v in vs]))
                    warn('cannot find a path that connects all the vertices of a cell', RuntimeWarning)
                    if try_fix_corners:
                        vs = vertices[cell_vertices[i]]
                        if len(vs) != 3:
                            _min = vs.min(axis=0) == np.r_[xlim[0],ylim[0]]
                            _max = vs.max(axis=0) == np.r_[xlim[1],ylim[1]]
                            if _min[0]:
                                vx = v0x = xlim[0]
                                v1x = vs[:,0].max()
                            elif _max[0]:
                                vx = v0x = xlim[1]
                                v1x = vs[:,0].min()
                            else:
                                vx = None
                            if _min[1]:
                                vy = v1y = ylim[0]
                                v0y = vs[:,1].max()
                            elif _max[1]:
                                vy = v1y = ylim[1]
                                v0y = vs[:,1].min()
                            else:
                                vy = None
                            if vx is None or vy is None:
                                vs = None
                            else:
                                vs = np.array([[v0x,v0y],[vx,vy],[v1x,v1y]])
                        if vs is not None:
                            polygons.append((vs[:,0],vs[:,1]))
                            ids.append(i)
                    break
            v = ws.pop()
        #
        if _vertices:
            _vertices = np.vstack(_vertices)
            polygons.append((_vertices[:,0], _vertices[:,1]))
            ids.append(i)

    if not ids:
        ids = ix[ok]
    scalar_map = values.loc[ids].values

    if clim is None:
        vmin, vmax = scalar_map.min(), scalar_map.max()
        clim = {}
    else:
        vmin, vmax = clim
        clim = dict(vmin=vmin, vmax=vmax)
    scalar_map = mpl.colors.Normalize(**clim)(scalar_map)

    if colormap:
        color_map = mpl.cm.get_cmap(colormap)
    else:
        color_map = mpl.cm.viridis
    colors = [
            "#%02x%02x%02x" % (int(r), int(g), int(b)) for r, g, b, _ in 255*color_map(scalar_map)
            ]
    patch_kwargs = dict(fill_color=colors, line_width=0)
    glyph_renderers.append(
            figure.patches(*zip(*polygons), **patch_kwargs)
        )

    if delaunay or isinstance(delaunay, dict):
        if not isinstance(delaunay, dict):
            delaunay = {}
        glyph_renderers.append(
                plot_delaunay(cells, figure=figure, **delaunay)
            )

    figure.x_range = Range1d(*xlim)
    figure.y_range = Range1d(*ylim)

    if colorbar:
        low = clim.get('vmin', vmin)
        high = clim.get('vmax', vmax)
        if False:
            add_colorbar(figure, colormap, low, high, unit, clabel, colorbar_figure)
        else:
            update_colorbar(figure, low, high, unit, clabel, colorbar_figure)

    #plt.show(figure)
    return glyph_renderers


def add_colorbar(figure, colormap=None, low=0, high=1, unit=None, clabel=None, colorbar_figure=None):
    if colormap is None:
        color_map = 'Viridis256'
    elif isinstance(colormap, str):
        if colormap.lower() in ('greys','inferno','magma','plasma','viridis','cividis','turbo'):
            color_map = colormap[0].upper()+colormap[1:].lower()+'256'
        else:
            color_map = mpl.cm.get_cmap(colormap)
            color_map = [
                    "#%02x%02x%02x" % (int(r), int(g), int(b)) \
                    for r, g, b, _ in 255*color_map(np.linspace(0, 1, 256))
                    ]
    else:
        color_map = colormap
    try:
        color_map = LinearColorMapper(palette=color_map, low=low, high=high,
                name='color_mapper')
    except ValueError as e:
        try:
            import colorcet
        except ImportError:
            raise ValueError('colormap not found; try installing the colorcet package') from e
        else:
            raise
    color_bar = ColorBar(color_mapper=color_map, ticker=BasicTicker(name='ticker'),
            border_line_color=None, margin=0, name='color_bar')
    color_bar.background_fill_color = None
    #glyph_renderers.append(color_bar)
    if unit is None:
        unit = clabel
    if colorbar_figure is None:
        if unit is not None:
            color_bar.title = unit
        figure.add_layout(color_bar, place='right')
    else:
        if unit is not None:
            colorbar_figure.title.text = unit
            colorbar_figure.title_location = 'right'
            colorbar_figure.title.align = 'center'
        color_bar.height = colorbar_figure.plot_height
        if isinstance(colorbar_figure.center[-1], ColorBar):
            colorbar_figure.center = colorbar_figure.center[:-1]
        colorbar_figure.add_layout(color_bar, place='center')
        #print(colorbar_figure.center, colorbar_figure.plot_height, color_bar.width, color_bar.height, color_bar.margin, color_bar.padding)


def update_colorbar(figure, low=None, high=None, unit=None, clabel=None, colorbar_figure=None):
    if colorbar_figure is not None:
        figure = colorbar_figure
    color_bar = figure.select(name='color_bar')[0]
    if not (low is None and high is None):
        color_map = figure.select(name='color_mapper')[0]
        palette = color_map.palette
        #color_map.update(low=low, high=high)
        #palette = 'Greys256'
        color_map = LinearColorMapper(palette=palette, low=low, high=high,
                name='color_mapper')
        color_bar.update(color_mapper=color_map)
    if unit is None:
        unit = clabel
    if colorbar_figure is None:
        if unit is not None:
            color_bar.title = unit
    else:
        if unit is not None:
            colorbar_figure.title.text = unit
        #color_bar.height = colorbar_figure.plot_height
        #if isinstance(colorbar_figure.center[-1], ColorBar):
        #    colorbar_figure.center = colorbar_figure.center[:-1]


def plot_delaunay(cells, labels=None, color=None, style='-',
        figure=None, linewidth=1, **kwargs):
    """
    Delaunay plot.

    Arguments:

        cells (Partition):
            full partition

        labels (numpy.ndarray):
            numerical labels for cell adjacency relationship

        color (str):
            single-character colours in a string, e.g. 'rrrbgy'

        style (str):
            line style

        figure (bokeh.plotting.figure.Figure):
            figure handle

        linewidth (int):
            line width

    Returns:

        tuple: list of handles of the plotted edges,
            handle of the plotted centroids
    """
    if figure is None:
        raise NotImplementedError
    try:
        tessellation = cells.tessellation
    except AttributeError:
        tessellation = cells

    vertices = tessellation.cell_centers

    labels, color = _graph_theme(tessellation, labels, color, True)

    # if asymmetric, can be either triu or tril
    A = sparse.triu(tessellation.cell_adjacency, format='coo')
    I, J, K = A.row, A.col, A.data
    if not I.size:
        A = sparse.tril(tessellation.cell_adjacency, format='coo')
        I, J, K = A.row, A.col, A.data

    by_color = defaultdict(list)
    edge_handles = []

    # plot delaunay
    for i, j, k in zip(I, J, K):
        x, y = zip(vertices[i], vertices[j])
        if labels is None:
            c = 0
        else:
            label = tessellation.adjacency_label[k]
            try:
                c = labels.index(label)
            except ValueError:
                continue
        by_color[c].append((x, y))

    for c in by_color:
        xy = by_color[c]
        X = np.zeros((len(xy) * 3,))
        Y = np.empty((len(xy) * 3,))
        Y[:] = np.nan
        i = 0
        for x, y in xy:
            I = slice(i*3, i*3+2)
            X[I], Y[I] = x, y
            i += 1
        h = figure.line(X, Y, line_dash=_line_style_to_dash_pattern(style),
            line_color=long_colour_name(color[c if color[1:] else 0]),
            line_width=linewidth)
        edge_handles.append(h)

    return edge_handles


def _line_style_to_dash_pattern(style):
    return {
        '-':    [],
        '-.':   'dashdot',
        ':':    'dotted',
        '.-':   'dotdash',
        '--':   'dashed',
        }.get(style, style)


def field_map_2d(cells, values, angular_width=30.0,
        figure=None,
        cell_arrow_ratio=None,
        markercolor='#bfbf00', markeredgecolor='#000000', markeralpha=0.8, markerlinewidth=None,
        inferencemap=False,
        **kwargs):
    """
    Plot a 2D field (vector) map as arrows.

    Arguments:

        cells (Partition or Distributed): spatial description of the cells

        values (pandas.DataFrame or numpy.ndarray): value at each cell, represented as a colour

        angular_width (float): angle of the tip of the arrows

        figure (matplotlib.figure.Figure): figure handle

        cell_arrow_ratio (float): size of the largest arrow relative to the median
            inter-cell distance; default is ``0.4`` if `inferencemap` is ``True``,
            else ``1``

        markercolor (str): colour of the arrows

        markeredgecolor (str): colour of the border of the arrows

        markeralpha (float): alpha value of the arrows

        markerlinewidth (float): line width of the border of the arrows

        inferencemap (bool): if ``True``, the arrow length only depends on the cell size

    """
    force_amplitude = values.pow(2).sum(1).apply(np.sqrt)
    if figure is None:
        figure = plt.figure()
    if markercolor is None and 'color' in kwargs:
        markercolor = kwargs['color']
    if markeredgecolor is None and 'edgecolor' in kwargs:
        markeredgecolor = kwargs['edgecolor']
    if markeralpha is None and 'alpha' in kwargs:
        markeralpha = kwargs['alpha']
    if markerlinewidth is None and 'linewidth' in kwargs:
        markerlinewidth = kwargs['linewidth']
    xy = cells.tessellation.cell_centers
    xy_min, _, xy_max, _ = mplt._bounding_box(cells, xy)
    xlim = (xy_min[0], xy_max[0])
    ylim = (xy_min[1], xy_max[1])
    xmin, xmax = xlim
    ymin, ymax = ylim
    # identify the visible cell centers
    if isinstance(cells, Tessellation):
        pts = cells.cell_centers
    elif isinstance(cells, FiniteElements):
        pts = np.vstack([ cells[i].center for i in cells ])#values.index ])
    elif isinstance(cells, Partition):
        assert isinstance(cells.tessellation, Tessellation)
        pts = cells.tessellation.cell_centers#[values.index]
    inside = (xmin<=pts[:,0]) & (pts[:,0]<=xmax) & (ymin<=pts[:,1]) & (pts[:,1]<=ymax)
    # compute the distance between adjacent cell centers
    if isinstance(cells, Delaunay):
        A = cells.cell_adjacency
    elif isinstance(cells, FiniteElements):
        A = cells.adjacency
    elif isinstance(cells, Partition):
        if isinstance(cells.tessellation, Delaunay):
            A = cells.tessellation.cell_adjacency
        else:
            raise TypeError('unsupported tessellation type: {}'.format(type(cells.tessellation)))
    else:
        raise TypeError('unsupported cell type: {}'.format(type(cells)))
    if inferencemap:
        if cell_arrow_ratio is None:
            cell_arrow_ratio = .5
        A = A.tocsr()
        scale = np.full(A.shape[0], np.nan, dtype=pts.dtype)
        for i in values.index:
            if not inside[i]:
                continue
            js = A.indices[A.indptr[i]:A.indptr[i+1]]
            pts_i, pts_j = pts[[i]], pts[js[inside[js]]]
            if pts_j.size:
                inter_cell_distance = pts_i - pts_j
                inter_cell_distance = np.sqrt(np.sum(inter_cell_distance * inter_cell_distance, axis=1))
                #scale[i] = np.nanmean(inter_cell_distance)
                scale = np.nanquantile(inter_cell_distance, .1) * cell_arrow_ratio
    else:
        if cell_arrow_ratio is None:
            cell_arrow_ratio = .4 # backward compatibility
        A = sparse.triu(A, format='coo')
        I, J = A.row, A.col
        _inside = inside[I] & inside[J]
        pts_i, pts_j = pts[I[_inside]], pts[J[_inside]]
        inter_cell_distance = pts_i - pts_j
        inter_cell_distance = np.sqrt(np.sum(inter_cell_distance * inter_cell_distance, axis=1))
        # scale force amplitude
        large_arrow_length = np.nanmax(force_amplitude[inside[values.index]]) # TODO: clipping
        scale = np.nanmedian(inter_cell_distance) / (large_arrow_length * cell_arrow_ratio)
    #
    dw = float(angular_width) / 2.0
    t = tan(radians(dw))
    t = np.array([[0.0, -t], [t, 0.0]])
    markers = []
    for i in values.index:
        center = pts[i]
        radius = force_amplitude[i]
        if inferencemap:
            f = np.asarray(values.loc[i])
            f *= scale / max(1e-16, np.sqrt(np.sum(f * f)))
        else:
            f = np.asarray(values.loc[i]) * scale
        base, vertex = center + np.outer([-1./3, 2./3], f)
        ortho = np.dot(t, f)
        vertices = np.stack((vertex, base + ortho, base - ortho), axis=0)
        #vertices[:,0] = center[0] + aspect_ratio * (vertices[:,0] - center[0])
        markers.append((vertices[:,0], vertices[:,1]))

    patch_kwargs = dict(fill_color=markercolor, fill_alpha=markeralpha,
            line_width=markerlinewidth, line_color=markeredgecolor)
    return figure.patches(*zip(*markers), **patch_kwargs)



__all__ = ['plot_points', 'plot_trajectories', 'plot_delaunay', 'scalar_map_2d', 'field_map_2d',
        'add_colorbar']





from ..attribute import single
from ..artefact  import Analysis
from ..roi       import ROI
from collections import OrderedDict
import numpy  as np
import pandas as pd


def as_time_profiles(self, sampling, maps, space_bin_index=None, segment_indices=None,
        return_times=False, return_bin_index=False):
    if isinstance(sampling, ROI):
        roi = sampling
        sampling = roi.get_sampling()
    if isinstance(maps, (int, str)):
        if isinstance(sampling, Analysis):
            map_label = maps
            maps = sampling.get_child(map_label)
    times = {}
    profiles = OrderedDict()
    for _s, _t, _sampling, _maps in self.time.as_time_segments(sampling, maps,
            index=segment_indices, return_index=True, return_times=True):
        if space_bin_index is None:
            space_bin_index = _maps.index
        elif np.isscalar(space_bin_index):
            space_bin_index = [ space_bin_index ]
        for b in space_bin_index:
            if b in _maps.index:
                _values = _maps.loc[[b]].copy()
                _values.index = [ _s ]
                if b not in profiles:
                    times[b] = []
                    profiles[b] = []
                times[b].append(_t)
                profiles[b].append(_values)
    for b in profiles:
        profile = pd.concat(profiles[b])
        ret = []
        if return_bin_index:
            ret.append(b)
        if return_times:
            ts = np.stack(times[b], axis=0)
            ret.append(ts)
        ret.append(profile)
        yield tuple(ret)

def get_time_profile(self, sampling, maps, space_bin_index, segment_indices=None,
        return_times=False):
    return single(as_time_profiles(self, sampling, maps, space_bin_index, segment_indices, return_times, False))


__all__ = ['as_time_profiles', 'get_time_profile']




# -*- coding: utf-8 -*-

# Copyright © 2017, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from __future__ import print_function
import importlib
import copy
import os
import re
try:
    fullmatch = re.fullmatch
except AttributeError: # Py2
    fullmatch = re.match
from warnings import warn
import traceback


def list_plugins(dirname, package, lookup={}, force=False, require=None, verbose=False):
    if verbose:
        _pre = 'loading: '
        _post = '...'
        _success = '[done]'
        _failure = '[failed]'
    if not require:
        require = ()
    elif isinstance(require, str):
        require = (require,)
    pattern = re.compile(r'[a-zA-Z0-9].*([.]py)?')
    candidate_modules = set([ os.path.splitext(fn)[0] \
        for fn in os.listdir(dirname) \
        if fullmatch(pattern, fn) is not None ])
    modules = {}
    provided = {}
    for name in candidate_modules:
        path = '{}.{}'.format(package, name)
        # load module
        try:
            module = importlib.import_module(path)
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            if verbose:
                print('{}{}{}\t{}'.format(_pre, path, _post, _failure))
                print(traceback.format_exc(), end='')
            continue
        # ensure that all the required attributes are available
        _continue = False
        for required in require:
            if not hasattr(module, required):
                _continue = True
                break
        if _continue:   continue
        elif verbose:
            print('{}{}{}'.format(_pre, path, _post), end='\t')
        # parse setup
        if hasattr(module, 'setup'):
            setup = module.setup
            try:
                name = setup['name']
            except KeyError:
                setup['name'] = name
        else:
            setup = dict(name=name)
        # parse other attributes
        try:
            namespace = module.__all__
        except AttributeError:
            namespace = list(module.__dict__.keys())
        missing = conflicting = None
        warning = []
        for key in lookup:
            if key in setup:
                continue
            ref = lookup[key]
            if isinstance(ref, type):
                matches = []
                for var in namespace:
                    try:
                        ok = issubclass(getattr(module, var), ref)
                    except TypeError:
                        ok = False
                    if ok:
                        matches.append(var)
            else:
                matches = [ var for var in namespace
                    if fullmatch(ref, var) is not None ]
            if matches:
                if matches[1:]:
                    conflicting = key
                    if not force:
                        break
                setup[key] = matches[0]
            else:
                missing = key
                if not force:
                    break
        if conflicting:
            warning.append(("multiple matches in module '{}' for key '{}'".format(path, conflicting), ImportWarning))
            if not force:
                continue
        if missing:
            warning.append(("no match in module '{}' for key '{}'".format(path, missing), ImportWarning))
            if verbose:
                if force:
                    print(_success)
                else:
                    print(_failure)
        elif verbose:
            print(_success)
        for w in warning:
            warn(*w)
        if missing and not force:
            continue
        # register plugin
        plugin = (setup, module)
        if isinstance(name, str):
            modules[name] = plugin
        else:
            names = name
            for name in names:
                modules[name] = plugin
        try:
            provides = setup['provides']
        except KeyError:
            pass
        else:
            if isinstance(provides, str):
                provided[provides] = plugin
            else:
                for _provides in provides:
                    provided[_provides] = plugin
    for provides in provided:
        if provides not in modules:
            modules[provides] = provided[provides]
    return modules


def add_arguments(parser, arguments, name=None):
    translations = []
    for arg, options in arguments.items():
        if not options:
            continue
        long_arg = '--' + arg.replace('_', '-')
        has_options = False
        if isinstance(options, (tuple, list)):
            args = list(options)
            kwargs = args.pop()
            options = {}
        else:
            args = []
            try:
                _parse = options['parse']
            except KeyError:
                pass
            else:
                if callable(_parse):
                    translations.append((arg, _parse))
                    has_options = True
            try:
                kwargs = options['kwargs']
            except KeyError:
                if has_options or not options:
                    continue
                kwargs = options
                options = None # should not be used anymore
            else:
                has_options = True
                try:
                    args = list(options.get('args'))
                except KeyError:
                    pass
        if has_options and options.get('translate', False):
            try:
                _arg = args[1]
            except IndexError:
                _arg = args[0]
            _arg = _arg.lstrip('-').replace('-', '_')
            def _translate(**_kwargs):
                return _kwargs.get(_arg, None)
            translations.append((arg, _translate))
        elif long_arg not in args:
            if args:
                args.insert(1, long_arg)
            else:
                args = (long_arg,)
        try:
            parser.add_argument(*args, **kwargs)
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            if name:
                print("WARNING: option `{}` from plugin '{}' ignored".format(arg, name))
            else:
                print("WARNING: option `{}` ignored".format(arg))
    return translations


def short_options(arguments):
    _options = []
    for args in arguments.values():
        if args:
            if isinstance(args, (tuple, list)):
                args = args[0]
            elif 'args' in args:
                args = args['args']
            else:
                continue
            if not isinstance(args, (tuple, list)):
                args = (args,)
            for arg in args:
                if arg[0] != arg[1]:
                    _options.append(arg)
    return _options


class Plugins(object):

    __slots__ = ('modules', 'dirname', 'package', 'lookup', 'force', 'require', 'verbose', 'post_load')

    def __init__(self, dirname, package, lookup={}, force=False, require=None, verbose=False):
        self.modules = None
        self.dirname = dirname
        self.package = package
        self.lookup = dict(lookup)
        self.force = force
        self.require = require
        self.verbose = verbose
        self.post_load = None

    def __load__(self):
        if self.modules is None:
            self.modules = list_plugins(
                self.dirname,
                self.package,
                self.lookup,
                self.force,
                self.require,
                self.verbose,
                )
            if self.post_load:
                self.post_load(self)

    def __repr__(self):
        return 'Plugins{{in \'{}\'}}{}'.format(self.package, '{(not loaded)}' if self.modules is None else repr(self.modules))

    def __nonzero__(self):
        self.__load__()
        return self.modules.__nonzero__()

    def __len__(self):
        self.__load__()
        return self.modules.__len__()

    def __iter__(self):
        self.__load__()
        return self.modules.__iter__()

    def __getitem__(self, mod):
        self.__load__()
        return self.modules.__getitem__(mod)

    def __setitem__(self, mod, descr):
        self.__load__()
        self.modules.__setitem__(mod, descr)

    def __missing__(self, mod):
        self.__load__()
        return self.modules.__missing__(mod)

    def keys(self):
        self.__load__()
        return self.modules.keys()

    def values(self):
        self.__load__()
        return self.modules.values()

    def items(self):
        self.__load__()
        return self.modules.items()

    def get(self, mod, default):
        self.__load__()
        return self.modules.get(mod, default)

    def pop(self, mod, default):
        self.__load__()
        return self.modules.pop(mod, default)

    def update(self, plugins):
        self.__load__()
        if not isinstance(plugins, dict):
            raise TypeError('not a `dict`')
        self.modules.update(plugins)


__all__ = [
    'list_plugins',
    'add_arguments',
    'short_options',
    'Plugins',
    ]




# -*- coding: utf-8 -*-

# Copyright © 2017-2021, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


import os
import numpy as np
import pandas as pd
import warnings
import itertools
from .exceptions import *
import re


def _translocations(df, sort=True): # very slow; may be marked as deprecated
    def diff(df):
        df = df.sort_values('t')
        t = df['t'][1:]
        df = df.diff()
        df['t'] = t
        return df
    return df.groupby(['n'], sort=False).apply(diff)

def translocations(df, sort=False):
    '''Each trajectories should be represented by consecutive rows sorted by time.

    Returns displacements without the associated position information.'''
    if sort:
        return _translocations(df) # not exactly equivalent
        #raise NotImplementedError
    i = 'n'
    xyz = ['x', 'y']
    if 'z' in df.columns:
        xyz.append('z')
    dxyz = [ 'd'+col for col in xyz ]
    if all( col in df.columns for col in dxyz ):
        jump = df[dxyz]
    else:
        ixyz = xyz + [i]
        jump = df[ixyz].diff()
        #df[xyz] = jump[xyz]
        #df = df[jump[i] != 0]
        #return df
        jump = jump[jump[i] == 0][xyz]
    return jump#np.sqrt(np.sum(jump * jump, axis=1))


def iter_trajectories(trajectories, trajnum_colname='n', asslice=False, asarray=False, order=None):
    """
    Yields the different trajectories in turn.

    If ``asslice=True``, the indices corresponding to the trajectory are returned instead,
    or as first output argument if ``asarray=True``, as a slice (first index, last index + 1).

    `order` can be *'start'* to ensure that trajectories are yielded by ascending start time.
    """
    if not isinstance(trajectories, pd.DataFrame):
        raise TypeError('trajectories is not a DataFrame')
    if trajectories.empty:
        return ()

    if asarray or not asslice:
        other_cols = [ col for col in trajectories.columns if col != trajnum_colname ]
    if asarray:
        dat = trajectories[other_cols].values
    if asslice:
        if asarray:
            from_slice = lambda a,b: (a,b), dat[a:b]
        else:
            from_slice = lambda a,b: (a,b)
    else:
        if asarray:
            from_slice = lambda a,b: dat[a:b]
        else:
            from_slice = lambda a,b: trajectories.iloc[a:b]

    if order is None:

        curr_traj_num = trajectories[trajnum_colname].iat[0] - 1
        for i, num in enumerate(trajectories[trajnum_colname]):
            if num == curr_traj_num:
                stop += 1
            elif num < curr_traj_num:
                raise IndexError('trajectories are not ascendingly sorted')
            else:
                if 0<i:
                    yield from_slice(start, stop)
                start = i
                stop = start + 1
                curr_traj_num = num
        yield from_slice(start, stop)

    elif order == 'start':

        new_n = 0<np.diff(trajectories['n'].values)
        new_n = np.flatnonzero(new_n)
        if new_n.size:
            new_n += 1
            traj_ids = np.stack((np.r_[0,new_n], np.r_[new_n,len(trajectories)]), axis=1)
            traj_ts = trajectories['t'].iloc[traj_ids[:,0]].values
            traj_ids = traj_ids[np.argsort(traj_ts)]
        else:
            traj_ids = np.array([[0,len(trajectories)]])
        for i,j in traj_ids:
            yield from_slice(i,j)


def iter_full_trajectories(cropped_trajs, all_trajs, match_cols=list('xyt'), unique=True):
    """
    In the case `cropped_trajs` results from cropping `all_trajs`,
    yields the trajectories in `all_trajs` that are in `cropped_trajs`.

    This function is helpful for retrieving the original trajectories,
    with the excluded points included back.
    Indeed, `crop` removes the out-of-bound locations, splits the affected
    trajectories and re-indices the trajectories, so that they are contiguous
    in time.

    There is no need to call `translocations_to_trajectories` on `cropped_trajs`
    beforehands.

    If `all_trajs` actually represent series of translocations,
    `iter_full_trajectories` yields full series of translocations.

    Argument `unique` set to :const:`False` makes `iter_full_trajectories`
    yield the trajectories that correspond to those yielded by `iter_trajectories`,
    so that both output can be zipped.
    The following code example iterates over regions of interest using the
    :class:`~tramway.analyzer.RWAnalyzer` object:

    .. code-block:: python

        from tramway.core.xyt import *
        from tramway.analyzer import *

        a = RWAnalyzer()

        # ... [define the SPT data and ROI]

        for f in a.spt_data:
            all_trajectories = f.dataframe
            for r in f.roi.as_support_regions():
                local_translocations = r.crop()
                local_trajectories = translocations_to_trajectories(local_translocations)
                for cropped_trajectory, full_trajectory in zip(
                        iter_trajectories(local_trajectories),
                        iter_full_trajectories(local_translocations, all_trajectories, unique=False),
                    ):
                    # do sommething with `cropped_trajectory` and corresponding `full_trajectory`
                    pass

    """
    visited_indices = set()
    all_coords = all_trajs[match_cols].values
    for traj in iter_trajectories(cropped_trajs):
        sample_loc = traj.iloc[0][match_cols].values
        matching_row_index = np.flatnonzero(np.all(sample_loc==all_coords, axis=1))
        assert len(matching_row_index)==1
        matching_index = all_trajs.iloc[matching_row_index[0]]['n']
        if not (unique and matching_index in visited_indices):
            visited_indices.add(matching_index)
            yield all_trajs[all_trajs['n']==matching_index]


def iter_frames(points, asslice=False, as_trajectory_slices=False, dt=None, skip_empty_frames=True):
    """
    Yields series of row indices, each series corresponding to a different frame.
    """
    if not isinstance(points, pd.DataFrame):
        raise TypeError('input data is not a DataFrame')
    ts = points['t'].values
    t = ts.min()
    if not np.isclose(t, ts[0]):
        raise ValueError('first location or trajectory starts later than t0')
    if 'n' in points:
        # trajectories or translocations; locations are grouped by trajectory
        if asslice:
            raise ValueError('rows are not contiguous in time; cannot represent frames as row slices')
        it = iter_trajectories(points, asslice=True, order='start')
        exhausted = False
        active_trajs = []
        while True:
            try:
                traj_ids = next(it)
            except StopIteration:
                exhausted = True
                next_traj = None
                break
            if np.isclose(ts[traj_ids[0]], t):
                active_trajs.append(traj_ids)
            else:
                next_traj = traj_ids
                break
        if dt is None:
            dt = np.median(np.diff(ts[traj_ids[0]:traj_ids[1]]))
        elif dt <= 0:
            raise ValueError('dt is not strictly positive')
        while True:
            frame = []
            trajs = []
            still_active = []
            for traj in active_trajs:
                traj_ts = ts[traj[0]:traj[1]]
                row_ix = int(np.round((t-traj_ts[0])/dt))
                if not np.isclose(traj_ts[row_ix], t):
                    traj_dt = np.diff(traj_ts)
                    ok = np.isclose(traj_dt , dt)
                    if np.any(ok) or len(traj)==2:
                        raise ValueError('a trajectory is not contiguous in time:\n{}'.format(points[traj[0]:traj[1]]))
                    else:
                        raise ValueError('dt may not be properly set: {}'.format(dt))
                frame.append(traj[0]+row_ix)
                trajs.append(traj)
                if t+dt/2<traj_ts[-1]:
                    still_active.append(traj)
            if frame:
                if as_trajectory_slices:
                    yield trajs
                else:
                    yield np.array(frame)
            else:
                raise RuntimeError('no trajectories found in frame')
            #
            active_trajs = still_active
            if active_trajs:
                t += dt
            elif exhausted:
                break
            else:
                next_t = ts[next_traj[0]]
                if not skip_empty_frames:
                    t += dt
                    while t<next_t-dt/2:
                        yield []
                        t += dt
                t = next_t
            #
            if not exhausted and np.isclose(ts[next_traj[0]], t):
                active_trajs.append(next_traj)
                while True:
                    try:
                        traj_ids = next(it)
                    except StopIteration:
                        exhausted = True
                        next_traj = None
                        break
                    if np.isclose(ts[traj_ids[0]], t):
                        active_trajs.append(traj_ids)
                    else:
                        next_traj = traj_ids
                        break
    else:
        if as_trajectory_slices:
            raise ValueError('input data are not trajectories')
        if not skip_empty_frames and (dt is None or dt == 0):
            raise ValueError('dt is undefined')
        if asslice:
            from_slice = lambda a,b: (a,b)
            empty_slice = ()
        else:
            from_slice = lambda a,b: np.arange(a,b)
            empty_slice = np.arange(0)

        i, t_i, j = 0, t, 1
        for t_j in ts:
            if np.issclose(t_j, t_i):
                j += 1
            else:
                yield from_slice(i, j)
                if not skip_empty_frames:
                    t_i += dt
                    while not np.isclose(t_j, t_i):
                        yield empty_slice
                        t_i += dt
                i, t_i = j, t_j
        if i<j:
            yield from_slice(i, j)


def load_xyt(path, columns=None, concat=True, return_paths=False, verbose=False,
        reset_origin=False, header=None, **kwargs):
    """
    Load trajectory files.

    Files are loaded with :func:`~pandas.read_csv` and should have the same number of columns
    and either none or all files should exhibit a single-line header.

    Default column names are 'n', 'x', 'y' and 't'.

    Arguments:

        path (str or list of str): path to trajectory file or directory.

        columns (list of str): column names.

        concat (bool): if multiple files are read, return a single DataFrame.

        return_paths (bool): paths to files are returned as second output argument.

        verbose (bool): print extra messages.

        reset_origin (bool or sequence): the lowest coordinate is translated to 0.
            Apply to time and space columns. Default column names are 'x', 'y', 'z'
            and 't'. A sequence overrides the default.

        header (bool): if defined, a single-line header is expected in the file(s);
            if ``False``, ignore the header;
            if ``True``, overwrite the `columns` argument with names from the header;
            if undefined, check whether a header is present and, if so, act as ``True``.

    Returns:

        pandas.DataFrame or list or tuple: trajectories as one or multiple DataFrames;
            if `tuple` (with *return_paths*), the trajectories are first, the list
            of filepaths second.

    Extra keyword arguments are passed to :func:`~pandas.read_csv`.
    """
    if columns is not None and header is True:
        raise ValueError('both column names and header are defined')
    #if 'n' not in columns:
    #    raise ValueError("trajectory index should be denoted 'n'")
    if 'sep' not in kwargs and 'delimiter' not in kwargs:
        kwargs['delim_whitespace'] = True
    if not isinstance(path, list):
        path = [path]
    paths = []
    for p in path:
        if os.path.isdir(p):
            paths.append([ os.path.join(p, f) for f in os.listdir(p) ])
        else:
            paths.append([p])
    index_max = 0
    df = []
    paths = list(itertools.chain(*paths))
    if not paths:
        if verbose:
            print('nothing to load')
        return
    _failed = []
    for f in paths:
        try:
            if verbose:
                print('loading file: {}'.format(f))
            if header is False:
                if columns is None:
                    columns = ['n', 'x', 'y', 't']
                kwargs['names'] = columns
                dff = pd.read_csv(f, header=0, **kwargs)
            else:
                with open(f, 'r') as fd:
                    first_line = fd.readline()
                if re.search(r'[a-df-zA-DF-Z_]', first_line):
                    if columns is None:
                        columns = first_line.split()
                    kwargs['names'] = columns
                    dff = pd.read_csv(f, header=0, **kwargs)
                elif header is True:
                    dff = pd.read_csv(f, header=0, **kwargs)
                    columns = dff.columns
                else:
                    if columns is None:
                        columns = ['n', 'x', 'y', 't']
                    kwargs['names'] = columns
                    dff = pd.read_csv(f, **kwargs)
        except OSError:
            _failed.append(f)
        else:
            if 'n' in columns:
                sample = dff[dff['n']==dff['n'].iloc[-1]]
                sample_dt = sample['t'].diff()[1:]
                if not all(0 < sample_dt):
                    if any(0 == sample_dt):
                        try:
                            conflicting = sample_dt.values == 0
                            conflicting = np.logical_or(np.r_[False, conflicting], np.r_[conflicting, False])
                            print(sample.loc[conflicting])
                        except:
                            pass
                        raise ValueError("some simultaneous locations are associated to a same trajectory: '{}'".format(f))
                    else:
                        warnings.warn(EfficiencyWarning("table '{}' is not properly ordered".format(f)))
                    # faster sort
                    data = np.asarray(dff)
                    dff = pd.DataFrame(data=data[np.lexsort((dff['t'], dff['n']))],
                        columns=dff.columns)
                    #sorted_dff = []
                    #for n in dff['n'].unique():
                    #       sorted_dff.append(dff[dff['n'] == n].sort_values(by='t'))
                    #dff = pd.concat(sorted_dff)
                    #dff.index = np.arange(dff.shape[0]) # optional
                if dff['n'].min() < index_max:
                    dff['n'] += index_max
                    index_max = dff['n'].max()
            undefined = dff.isnull().values.all(axis=0)
            if np.any(undefined):
                if columns == list('nxyt') and np.sum(undefined) == 1:
                    raise ValueError('the molecules are not tracked')
                else:
                    raise ValueError('too many specified columns: {}'.format(columns))
            df.append(dff)
    if df:
        for f in _failed:
            warnings.warn(f, FileNotFoundWarning)
    else:
        if paths[1:]:
            raise OSError('cannot load any of the files: {}'.format(paths))
        else:
            raise OSError('cannot load file: {}'.format(paths[0]))
    if reset_origin:
        if reset_origin == True:
            reset_origin = [ col for col in ['x', 'y', 'z', 't'] if col in columns ]
        origin = dff[reset_origin].min().values
        for dff in df[:-1]:
            origin = np.minimum(origin, dff[reset_origin].min().values)
        for dff in df:
            dff[reset_origin] -= origin
    if concat:
        if df[1:]:
            df = pd.concat(df)
        else:
            df = df[0]
    if return_paths:
        return (df, paths)
    else:
        return df


def crop(points, box, by=None, add_deltas=True, keep_nans=False, no_deltas=False, keep_nan=None,
        preserve_index=False):
    """
    Remove locations outside a bounding box.

    When a location is discarded, the corresponding trajectory is splitted into two distinct
    trajectories.

    Important: the locations of any given trajectory should be contiguous and ordered.

    Arguments:

        locations (pandas.DataFrame): locations with trajectory indices in column 'n',
            times in column 't' and coordinates in the other columns;
            delta columns are ignored

        box (array-like): origin and size of the space bounding box

        by (str): for translocations only;
            '*start*' or '*origin*': crop by translocation origin; keep the associated destinations;
            '*stop*' or '*destination*': crop by translocation destinations; keep the associated
            origins;
            trajectories with a single non-terminal point outside the bounding box are
            not splitted

        add_deltas (bool): add 'dx', 'dy', ..., 'dt' columns is they are not already present;
            deltas are associated to the translocation origins

        keep_nans/keep_nan (bool): adding deltas generates NaN; keep them

        no_deltas (bool): do not consider any column as deltas

        preserve_index (bool): do not split the trajectories with out-of-bound locations,
            do not re-index the trajectories nor the rows

    Returns:

        pandas.DataFrame: filtered (trans-)locations
    """
    box = np.asarray(box)
    dim = int(box.size / 2)
    support_lower_bound = box[:dim]
    support_size = box[dim:]
    support_upper_bound = support_lower_bound + support_size
    not_coord_cols = ['n', 't']
    if no_deltas:
        delta_cols = []
    else:
        delta_cols = [ c for c in points.columns \
                if c[0]=='d' and c[1:] != 'n' and c[1:] in points.columns ]
        not_coord_cols += delta_cols
    coord_cols = [ c for c in points.columns if c not in not_coord_cols ]
    if len(coord_cols) != dim:
        raise ValueError('the bounding box has dimension {} while the following coordinate columns were found: {}'.format(dim, coord_cols))
    within = np.all(np.logical_and(support_lower_bound <= points[coord_cols].values,
        points[coord_cols].values <= support_upper_bound), axis=1)
    if add_deltas or by:
        paired_dest = points['n'].diff().values==0
        paired_src = np.r_[paired_dest[1:], False]
    points = points.copy()
    if add_deltas:
        cols_with_deltas = [ c[1:] for c in delta_cols ]
        cols_to_diff = [ c for c in points.columns if c not in ['n']+cols_with_deltas+delta_cols ]
        deltas = points[cols_to_diff].diff().shift(-1)
        deltas = deltas[paired_src]
        deltas.columns = [ 'd'+c for c in cols_to_diff ]
        points = points.join(deltas)
    if by:
        if by in ('start', 'origin'):
            within[paired_dest] |= within[paired_src]
        elif by in ('stop', 'destination'):
            within[paired_src] |= within[paired_dest]
        else:
            raise ValueError('unsupported value for argument `by`')
    if 'n' in points.columns:
        n = points['n'] + np.cumsum(np.logical_not(within), dtype=points.dtypes['n'])
        single_point = 0 < n.diff().values[1:]
        single_point[:-1] = np.logical_and(single_point[:-1], single_point[1:])
        ok = np.r_[True, np.logical_not(single_point)]
        points = points[ok]
        within = within[ok]
        if not preserve_index:
            n = n[ok]
            n -= (n.diff().fillna(0).astype(int) - 1).clip(lower=0).cumsum()
            points['n'] = n
    points = points[within]
    if keep_nan is None:
        keep_nan = keep_nans
    if not keep_nan:
        points.dropna(inplace=True)
    if not preserve_index:
        points.index = np.arange(points.shape[0])
    return points


def reindex_trajectories(trajectories, trajnum_colname='n', dt=None):
    """
    Splits the trajectories with missing time steps and assigns different indices to
    the different segments.

    Works with trajectories and translocations.
    """
    are_translocations = 'dt' in trajectories.columns and not np.any(np.isnan(trajectories['dt']))
    trajs = []
    for traj in iter_trajectories(trajectories, trajnum_colname):
        if dt is None:
            if are_translocations:
                dt = traj['dt'].min()
            else:
                dt = traj['t'].diff().min()
            if dt.size == 0:
                dt = None
            elif np.isclose(dt, 0):
                raise ValueError('multiple rows for the same time point and trajectory')
        if dt is None:
            holes = np.array([])
        elif are_translocations:
            holes = 1.1 * dt <= traj['dt'].values
            holes[-1] = False
            holes = np.flatnonzero(holes)
        else:
            frame_ids = np.round(traj['t'].values/dt)
            steps = np.diff(frame_ids)
            holes = np.flatnonzero(1 < steps)
        if holes.size:
            holes += 1
            i = 0
            for j in holes:
                trajs.append(j-i)
                i = j
            trajs.append(len(traj)-i)
        else:
            trajs.append(len(traj))
    n = np.repeat(np.arange(1,len(trajs)+1), trajs)
    trajs = trajectories.copy()
    trajs[trajnum_colname] = n
    return trajs


def discard_static_trajectories(trajectories, min_msd=None, trajnum_colname='n', full_trajectory=False, verbose=False, localization_error=None):
    """
    Arguments:

        trajectories (DataFrame): trajectory or translocation data with
            columns :const:`'n'` (trajectory number),
            spatial coordinates :const:`'x'` and :const:`'y'` (and optionaly
            :const:`'z'`), and time :const:`'t'`;
            delta columns, if available (translocations), are used instead
            for calculating the displacement length.

        min_msd (float): minimum mean-square-displacement (usually set to the localization error).

        trajnum_colname (str): column name for the trajectory number.

        full_trajectory (float or bool): if :const:`True`, the trajectories with
            static translocations are entirely discarded;
            if False, only the static translocations are discarded, and the
            corresponding trajectories are discarded only if they end up being
            single points;
            if a `float` value, trajectories with `full_trajectory` x 100% static
            translocations or more are discarded.

        localization_error (float): alias for `min_msd`; for backward compatibility.

    Returns:

        DataFrame: filtered trajectory data with a new row index.

    """
    if min_msd is None:
        min_msd = localization_error
    trajs = []
    for start,stop in iter_trajectories(trajectories, trajnum_colname, asslice=True, order='start'):
        traj = trajectories.iloc[start:stop]
        delta_cols = [ col for col in trajectories.columns if col.startswith('d') and col[1:] in trajectories.columns ]
        columns_with_deltas = [ col[1:] for col in delta_cols ]
        if columns_with_deltas:
            dr = traj[[ 'd'+col for col in trajectories.columns if col in 'xyz' ]].values
        else:
            r = traj[[ col for col in trajectories.columns if col in 'xyz' ]].values
            dr = np.diff(r, axis=0)
        js = np.mean(dr * dr, axis=1)
        if columns_with_deltas:
            js[np.isnan(js)] = -1
            static = js < min_msd
        else:
            static = np.r_[False, js < min_msd]
        if verbose and np.any(static):
            print('trajectory {:.0f} exhibits static translocations'.format(traj[trajnum_colname].iat[0]))
        if full_trajectory:
            if isinstance(full_trajectory, bool):
                if np.any(static):
                    # discard the entire trajectory
                    continue
            else:
                threshold_ratio = full_trajectory
                if columns_with_delta:
                    ratio = np.sum(static)/static.size
                else:
                    ratio = (np.sum(static)-1)/(static.size-1)
                if threshold_ratio <= ratio:
                    # discard the entire trajectory
                    continue
        else:
            traj = traj.loc[~static]
        if (columns_with_deltas and 0<len(traj)) or 1<len(traj):
            trajs.append(traj)
    if trajs:
        return pd.concat(trajs, ignore_index=True) # preserve column types
    else:
        return pd.DataFrame([], columns=trajectories.columns)


def load_mat(path, columns=None, varname='plist', dt=None, coord_scale=None, pixel_size=None):
    """
    Load SPT data from MatLab V7 file.

    The two pieces of code below are almost equivalent:

    .. code-block:: python

        xyt      = load_mat(my_file, columns=list('txy'), dt=frame_interval)

    .. code-block:: python

        xyt      = load_mat(my_file, columns=['frame_index', 'x', 'y'])
        xyt['t'] = xyt['frame_index'] * frame_interval

    The only difference resides in the :const:`'frame_index'` column
    that is missing in the first dataframe.

    Arguments:

        path (str): file path.

        columns (sequence of str): data column names;
            default is ['t', 'x', 'y'].

        varname (str): record name.

        dt (float): frame interval in seconds;
            if defined together with a 't' column in the data,
            the original 't' values are considered to be
            frame indices and are multiplied by dt to transform
            them into times.

        coord_scale (float): convertion factor for spatial coordinates.

        pixel_size (float): deprecated; superseded by `coord_scale`.

    Returns:

        pandas.DataFrame: SPT data.
    """
    import h5py
    try:
        with h5py.File(path, 'r') as f:
            spt_data = f[varname][...].T
    except OSError:
        from scipy.io import loadmat
        spt_data = loadmat(path)[varname]
    if columns is None:
        if spt_data.shape[0]==3:
            columns = list('txy')
        else:
            raise NotImplementedError('cannot infer the column names')
    spt_data = pd.DataFrame(spt_data, columns=columns)
    if dt is not None and 't' in spt_data.columns:
        spt_data['t'] = spt_data['t'] * dt
    if coord_scale is None:
        if pixel_size is not None:

            warnings.warn('attribute pixel_size is deprecated; use coord_scale instead', DeprecationWarning)
            coord_scale = pixel_size
    if coord_scale is not None:
        coords = [ c for c in ['x', 'y', 'z', 'dx', 'dy', 'dz'] if c in spt_data.columns ]
        spt_data[coords] = spt_data[coords] * coord_scale
    return spt_data


def trajectories_to_translocations(points, exclude_columns=['n']):
    """
    Appends delta columns (*'dx'*, *'dy'*, *'dt'*, etc) and removes
    the last location of each trajectory.

    See also `translocations_to_trajectories`.
    """
    exclude_columns = list(exclude_columns)
    delta_cols = [ c for c in points.columns \
            if c[0]=='d' and c[1:] not in exclude_columns and c[1:] in points.columns ]
    paired_dest = points['n'].diff().values==0
    paired_src = np.r_[paired_dest[1:], False]
    points = points.copy()
    cols_with_deltas = [ c[1:] for c in delta_cols ]
    cols_to_diff = [ c for c in points.columns if c not in exclude_columns+cols_with_deltas+delta_cols ]
    deltas = points[cols_to_diff].diff().shift(-1)
    deltas = deltas[paired_src]
    deltas.columns = [ 'd'+c for c in cols_to_diff ]
    points = points.join(deltas)
    points.dropna(inplace=True)
    points.index = np.arange(points.shape[0])
    return points

def translocations_to_trajectories(points):
    """
    Reintroduces the last location of each trajectory, and discards
    the delta columns.

    See also `trajectories_to_translocations`.
    """
    delta_cols = [ c for c in points.columns \
            if c[0]=='d' and c[1:] != 'n' and c[1:] in points.columns ]
    cols_with_deltas = [ c[1:] for c in delta_cols ]
    trans_n = points['n'].values
    traj_n = []
    traj_xyt = []
    r = 0
    while r<points.shape[0]:
        n = trans_n[r]
        start = r
        while r<points.shape[0] and trans_n[r]==n:
            r += 1
        stop = r
        trans = points.iloc[start:stop]
        traj = trans[cols_with_deltas].values
        traj_end = traj[-1] + trans[delta_cols].values[-1]
        traj_n.append(np.full((stop-start+1,1), n))
        traj_xyt.append(traj)
        traj_xyt.append(traj_end[np.newaxis,:])
    n = pd.DataFrame(np.vstack(traj_n), columns=['n'])
    xyt = pd.DataFrame(np.vstack(traj_xyt), columns=cols_with_deltas)
    return n.join(xyt)


__all__ = [
    'translocations',
    'iter_trajectories',
    'iter_full_trajectories',
    'iter_frames',
    'trajectories_to_translocations',
    'translocations_to_trajectories',
    'load_xyt',
    'load_mat',
    'crop',
    'discard_static_trajectories',
    'reindex_trajectories',
    ]




# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from abc import *
from ..attribute.abc import Attribute, abstractmethod

class SPTData(Attribute):
    """
    Abstract base class for the :attr:`~tramway.analyzer.RWAnalyzer.spt_data` attribute
    of an :class:`~tramway.analyzer.RWAnalyzer` object.
    """
    @property
    @abstractmethod
    def columns(self):
        """
        *list* of *str*: Data column names
        """
        pass
    @property
    @abstractmethod
    def localization_error(self):
        r"""
        *float*: Localization error in :math:`\mu m^2`
        """
        pass
    @property
    @abstractmethod
    def frame_interval(self):
        """
        *float*: Time interval between successive frames, in seconds
        """
        pass
    @abstractmethod
    def reset_origin(self, columns=None, same_origin=False):
        """ Translates the lower bound of the spatial data to zero.
        """
        pass
    @abstractmethod
    def discard_static_trajectories(self, dataframe=None, min_msd=None):
        """ The minimum mean-square-displacement is set to the localization error per default. 
        
        If an input dataframe is given, then :meth:`discard_static_trajectories` returns the processed
        dataframe and `self` does not maintain any copy of it.
        
        Otherwise, :meth:`discard_static_trajectories` applies inplace to the internal dataframe
        and does not return any dataframe.
        Note that it may set a flag instead of actually processing the internal dataframe,
        for the parcimonious processing in regions of interest only."""
        pass
    @abstractmethod
    def __iter__(self):
        """ Returns an iterator on the :class:`~tramway.analyzer.spt_data.abc.SPTDataItem`
        objects as natively stored. """
        pass
    @abstractmethod
    def as_dataframes(self, source=None):
        """ Generator function; yields :class:`pandas.DataFrame` objects.
        
        `source` can be a source name (filepath) or a boolean function
        that takes a source string as input argument."""
        if source is None:
            for f in self:
                with accessor(f, provides='dataframe') as a:
                    yield a.dataframe
        else:
            if callable(source):
                filter = source
            else:
                filter = lambda s: s == source
            for f in self:
                with accessor(f, provides=('source', 'dataframe')) as a:
                    if filter(a.source):
                        yield a.dataframe
    @property
    def bounds(self):
        """
        *pandas.DataFrame*: Lower (index *min*) and upper (index *max*) bounds for the SPT data
        """
        pass


class SPTDataItem(metaclass=ABCMeta):
    """
    Abstract base class for SPT data blocks an :class:`SPTData` object contains.
    """
    @property
    @abstractmethod
    def dataframe(self):
        """
        *pandas.DataFrame*: Raw SPT data
        """
        pass
    @property
    @abstractmethod
    def source(self):
        """
        *str*: File path or identifier
        """
        pass
    @property
    @abstractmethod
    def columns(self):
        pass
    columns.__doc__ = SPTData.columns.__doc__
    @property
    @abstractmethod
    def localization_error(self):
        pass
    localization_error.__doc__ = SPTData.localization_error.__doc__
    @property
    @abstractmethod
    def frame_interval(self):
        pass
    frame_interval.__doc__ = SPTData.frame_interval.__doc__
    @property
    @abstractmethod
    def bounds(self):
        pass
    bounds.__doc__ = SPTData.bounds.__doc__
    @property
    @abstractmethod
    def analyses(self):
        """
        *Analyses*: Analysis tree
        """
        pass

__all__ = [ 'SPTData', 'SPTDataItem' ]




from .rw_simulation import *
from .rw_features import *
from .batch_generation import *
from .batch_extraction import *
from .misc import *
try:
    from .vae import *
except ImportError: # torch
    pass
from .visualization import *
try:
    from .utils_supervised import *
except ImportError: # torch
    pass

"""
Using this module or subpackage is strongly discouraged,
as this code does not properly implement all of the featured measurements.
"""

import warnings
_msg = '''\
using the tramway.feature.single_traj module is strongly discouraged,
as this code does not properly implement all of the featured measurements\
'''
warnings.warn(_msg, FutureWarning)



# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


import tramway.localization.UNet.inference as deconv
import os.path
import numpy as np
from skimage import io


class _Files(object):
    def __init__(self, **kwargs):
        for key, val in kwargs.items():
            if val is None:
                setattr(self, key, val)
            else:
                attr = key[:-5] if key.endswith('_file') else key
                path = os.path.abspath(val) if isinstance(val, str) else val
                setattr(self, attr, path)


def deconvolve(image_stack_file, weight_file, mean_std_file=None,
        high_res_image_file=None, save_magnified_image=False,
        magnification=10, threshold=0, min_distance_peak=2, margin=3,
        header=True, abs_threshold=1., M=64, N=None, n=2, gpu=1):
    """
    """

    if N is None:
        N = M
    files = _Files(
            img_stack=image_stack_file,
            weights=weight_file,
            mean_std=mean_std_file,
            high_res_img=high_res_image_file,
            )

    if files.weights is None:
        weight_file = 'weight_model_one_GPU' if gpu==1 else 'weight_model'
        weight_file = os.path.normpath(os.path.join(os.path.dirname(__file__),
                '..', 'deconvolution', weight_file))
        files.weights = weight_file

    if files.mean_std is None:
        files.mean_std = os.path.normpath(os.path.join(os.path.dirname(__file__),
                '..', 'deconvolution', 'mean_std.txt'))
    mean_img, std_img = deconv.get_parameters_mean_std(files.mean_std)

    if save_magnified_image:
        deconv.save_trimmed_original_image_magnified_for_testing_purposes(
                files.img_stack, magnification, n)

    high_res_prediction, pos = deconv.Inference(files.img_stack,
            magnification, files.weights, mean_img, std_img, M, N, n,
            threshold, min_distance_peak, abs_threshold, margin, 1<gpu)

    basedir, filename = os.path.split(files.img_stack)
    basename,_ = os.path.splitext(filename)
    deconv.print_position_files(pos, basedir, basename, header)

    if files.high_res_img and high_res_prediction is not None:
        if not isinstance(files.high_res_img, str):
            files.high_res_img = os.path.join(basedir, 'predicted.tif')
        io.imsave(files.high_res_img, high_res_prediction.astype('uint16'))


def main():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('stack', help="path to the tiff image stack")
    parser.add_argument('--weights',  help="path to the weight file")
    parser.add_argument('--mean-std', help="path to the mean/std file (mean/std of the training images)")
    parser.add_argument('--magnification', default=10, type=int, help="magnification factor for input images")
    parser.add_argument('--gpu', type=int, default=1, help="number of GPUs")
    parser.add_argument('--disable-fixes', action='store_true', help="disable compatibility fixes for tensorflow<=1.14.0 and h5py>=3.0.0")
    args   = parser.parse_args()

    if args.disable_fixes:
        from tramway.localization.UNet import tf
        tf.__fix_tf_1_14_0_h5py_3_0_0__ = False

    deconvolve(args.stack, args.weights, args.mean_std, gpu=args.gpu,
            magnification=args.magnification)


if __name__ == '__main__':
    main()





import tempfile as tmp
import sys
import subprocess

def browse(files=[], browser='Firefox', colormap=None):
    if not files:
        import glob
        filepattern = '*.rwa'
        if not glob.glob(filepattern):
            filepattern = '*/*.rwa'
            if not glob.glob(filepattern):
                print('cannot find any matching rwa file')
                return
        files = [filepattern]

    script = tmp.NamedTemporaryFile('w', suffix='.py', encoding='utf-8', delete=False)
    source = """\
#!/usr/bin/env python3

from tramway.analyzer import *
from selenium import webdriver

a = RWAnalyzer()
a.spt_data.from_rwa_files(['{}'])
a.env.script = '{}'
{}a.browser.show_maps(webdriver=webdriver.{})
""".format("', '".join(files), script.name,
        '' if colormap is None else "a.browser.colormap = '{}'\n".format(colormap),
        browser)
    script.write(source)
    script.flush()
    p = subprocess.Popen([sys.executable, script.name],
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    try:
        ret = p.communicate()
    except KeyboardInterrupt:
        ret = p.communicate()
    script.close()
    def tostr(s):
        if isinstance(s, bytes):
            s = s.decode('utf8')
        return s
    ret = [ tostr(r) for r in ret ]
    return [ r for r in ret if r ]


def main():
    import argparse
    parser = argparse.ArgumentParser(prog='tramway-browse',
        description='Browse TRamWAy-generated .rwa files')
    parser.add_argument('files', nargs='*', help='for example: *.rwa or */*.rwa')
    parser.add_argument('--browser', default='Firefox', choices=['Firefox','Chrome','Edge','Ie','Opera','Safari','WebKitGTK'])
    parser.add_argument('--colormap', help="Matplotlib colormap name")
    [ print(ret) for ret in browse(**parser.parse_args().__dict__) ]

if __name__ == '__main__':
    main()





from .plugin import MapperPlugin
import numpy as np

class DV(MapperPlugin):
    __slots__ = ()
    def __init__(self, variant='standard', start=None, **kwargs):

        err = ValueError(f"not sure what to do with variant='{variant}' and start='{start}'")
        if variant in ('naive', 'original'):
            if start is not None:
                raise NotImplementedError(f"variant='standard' and start='{start}'")
            plugin = 'dv'
        elif variant.startswith('semi') and variant[5:] == 'stochastic':
            if not (start is None or start == 'stochastic'):
                raise NotImplementedError(f"variant='semi-stochastic' and start='{start}'")
            plugin = 'semi.stochastic.dv'
            start = 'stochastic'
        elif variant == 'stochastic':
            plugin = 'stochastic.dv'
            start = None
        elif variant == 'standard':
            if start == 'stochastic':
                plugin = 'semi.stochastic.dv'
            elif start is None:
                plugin = 'stochastic.dv'
            else:
                raise err
        else:
            raise err

        MapperPlugin.__init__(self, plugin, **kwargs)

        if variant == 'standard' and start is None:
            self.stochastic = False
            self.allow_negative_potential = True
        elif variant == 'naive':
            self.rgrad = 'delta0'
        elif plugin.endswith('stochastic.dv'):
            # as long as superlocal is True:
            self.ls_step_max = np.r_[.025, .5]
            self.ftol = 1e-3
            self.low_df_rate = 1
            self.allow_negative_potential = True
            if start == 'stochastic':
                self.max_iter = 20
        self.verbose = False

    @property
    def _rgrad_is_delta0(self):
        if self.rgrad is None:
            return self._name in ('stochastic.dv', 'semi.stochastic.dv')
        else:
            return self.rgrad == 'delta0'

    #@property
    #def diffusivity_prior(self):
    #    lambda_ = self.get_plugin_arg('diffusivity_prior')
    #    if self._rgrad_is_delta0 and lambda_ is not None:
    #        lambda_ = 2 * lambda_
    #    return lambda_

    #@diffusivity_prior.setter
    #def diffusivity_prior(self, lambda_):
    #    if self._rgrad_is_delta0 and lambda_ is not None:
    #        lambda_ = lambda_ / 2
    #    self.set_plugin_arg('diffusivity_prior', lambda_)

    #@property
    #def diffusivity_spatial_prior(self):
    #    lambda_ = self.get_plugin_arg('diffusivity_spatial_prior')
    #    if self._rgrad_is_delta0 and lambda_ is not None:
    #        lambda_ = 2 * lambda_
    #    return lambda_

    #@diffusivity_spatial_prior.setter
    #def diffusivity_spatial_prior(self, lambda_):
    #    if self._rgrad_is_delta0 and lambda_ is not None:
    #        lambda_ = lambda_ / 2
    #    self.set_plugin_arg('diffusivity_spatial_prior', lambda_)

    #@property
    #def potential_prior(self):
    #    lambda_ = self.get_plugin_arg('potential_prior')
    #    if self._rgrad_is_delta0 and lambda_ is not None:
    #        lambda_ = 2 * lambda_
    #    return lambda_

    #@potential_prior.setter
    #def potential_prior(self, lambda_):
    #    if self._rgrad_is_delta0 and lambda_ is not None:
    #        lambda_ = lambda_ / 2
    #    self.set_plugin_arg('potential_prior', lambda_)

    #@property
    #def potential_spatial_prior(self):
    #    lambda_ = self.get_plugin_arg('potential_spatial_prior')
    #    if self._rgrad_is_delta0 and lambda_ is not None:
    #        lambda_ = 2 * lambda_
    #    return lambda_

    #@potential_spatial_prior.setter
    #def potential_spatial_prior(self, lambda_):
    #    if self._rgrad_is_delta0 and lambda_ is not None:
    #        lambda_ = lambda_ / 2
    #    self.set_plugin_arg('potential_spatial_prior', lambda_)

    #@property
    #def verbose(self):
    #    return self.get_plugin_arg('verbose')

    #@verbose.setter
    #def verbose(self, v):
    #    import logging
    #    from tramway.inference.optimization import module_logger
    #    if v:
    #        lvl = logging.DEBUG
    #    else:
    #        v = True
    #        lvl = logging.WARNING
    #    module_logger.setLevel(lvl)
    #    self.set_plugin_arg('verbose', v)



# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from abc import *
from ..attribute.abc import Attribute, abstractmethod

class Images(Attribute):
    """
    Abstract base class for the :attr:`~tramway.analyzer.RWAnalyzer.images` attribute
    of an :class:`~tramway.analyzer.RWAnalyzer` object.
    """
    def as_frames(self, index=None, return_time=False):
        """ Generator function; yields frames as arrays.

        See for example :meth:`~tramway.analyzer.images._RawImage.as_frames`. """
        pass
    def crop_frames(self, bounding_box, index=None, return_time=False):
        """ Generator function similar to :meth:`as_frames`, with ROI cropping.

        See for example :meth:`~tramway.analyzer.images._RawImage.crop_frames`. """
        pass

class Image(metaclass=ABCMeta):
    """
    Abstract base class for single image stacks an :class:`Images` object contains.
    """
    pass


__all__ = ['Images', 'Image']




# -*- coding: utf-8 -*-

# Copyright © 2017, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


import matplotlib
#matplotlib.use('TkAgg')
#try:
#       import matplotlib.pyplot
#except: # TclError; head-less tramway
#       matplotlib.use('Agg')
#       # the present application will fail if run

from matplotlib.figure import Figure
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg

import numpy as np
import pandas as pd
try:
    from tramway.core.hdf5 import load_rwa, HDF5Store, peek_maps
except ImportError: # Sphinx?
    pass
import tramway.core.hdf5.compat
from tramway.core.analyses import *
from tramway.tessellation import Partition
from tramway.inference import Maps
from ..contour import ContourEditor
try:
    from tkinter import BooleanVar, IntVar, DoubleVar, StringVar
    import tkinter as tk
    import tkinter.filedialog as filedialog
    from tkinter import ttk
except ImportError: # Python 2
    from Tkinter import BooleanVar, IntVar, DoubleVar, StringVar
    import Tkinter as tk
    import tkFileDialog as filedialog
    import ttk
import os.path
import sys
import traceback



class TkContourEditor(FigureCanvasTkAgg):
    def __init__(self, parent, **kwargs):
        FigureCanvasTkAgg.__init__(self, Figure(), master=parent, **kwargs)
        self.plt = ContourEditor(self.figure, renderer=self)
        self.cell = IntVar()
        self.step = IntVar()
        self.variable = StringVar()
        self.integral = DoubleVar()
        self.cell.trace('w', self._cell)
        self.step.trace('w', self._step)
        self.variable.trace('w', self._variable)
        self.figure.canvas.mpl_connect('button_press_event', self.onclick)
        def _pass(*args, **kwargs):
            pass
        self.cell._report_exception = _pass
        self.step._report_exception = _pass
        self.plt.delaunay = True
        self.plt.callback = self.integral.set
        self.label = None

    def _cell(self, *args):
        c = self.cell.get()
        try:
            self.plt.cell = c
        except ValueError:
            self.cell.set(-1)
            raise
        except:
            print(traceback.format_exc())
            raise

    def _step(self, *args):
        s = self.step.get()
        if s < 1:
            self.step.set(1)
        else:
            try:
                self.plt.step = s
            except:
                print(traceback.format_exc())
                raise

    def _variable(self, *args):
        try:
            self.plt.variable = self.variable.get()
            if self.label is not None:
                if self.plt._variables[self.plt.variable][1:]:
                    self.label['text'] = 'Curl:'
                else:
                    self.label['text'] = 'Mean:'
        except:
            print(traceback.format_exc())
            raise

    def onclick(self, event):
        x, y = event.xdata, event.ydata
        if x is None or y is None:
            return
        self.cell.set(self.plt.find_cell((x, y)))

    def toggle_delaunay(self, *args):
        self.plt.delaunay = not self.plt._delaunay

    @property
    def cells(self):
        return self.plt.cells

    @cells.setter
    def cells(self, cs):
        try:
            self.plt.cells = cs
        except:
            print(traceback.format_exc())
            raise

    @property
    def map(self):
        return self.plt.map

    @map.setter
    def map(self, m):
        try:
            self.plt.map = m
        except:
            print(traceback.format_exc())
            raise

    @property
    def variables(self):
        return self.plt.variables

    @property
    def debug(self):
        return self.plt.debug

    @debug.setter
    def debug(self, d):
        self.plt.debug = d



class FileChooser(tk.Frame):
    def __init__(self, parent, filetypes={}, **kwargs):
        tk.Frame.__init__(self, parent, **kwargs)
        self.filetypes = dict(filetypes)
        self.filepath = StringVar()
        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)
        self.label = tk.Label(self, textvariable=self.filepath)
        self.label.grid(row=0, column=0, sticky=tk.W+tk.E)
        self.button = tk.Button(self, text='Open', command=self.open_input_file)
        self.button.grid(row=0, column=1, sticky=tk.E)

    def open_input_file(self):
        filepath = filedialog.askopenfilename(filetypes=self.filetypes)
        if filepath:
            self.filepath.set(filepath)



class ContourEditingApp(tk.Frame):
    def __init__(self, parent=None):
        tk.Frame.__init__(self, parent)
        self.analysis = StringVar()
        self.analysis.trace('w', self._analysis)
        self.analyses = None
        self.map_labels = []
        self.create_widgets()
        self.input_file.filepath.trace('w', self.new_file)
        self.editor.step.set(1)
        self.editor.label = self.integral_label

    def create_widgets(self):
        self.input_file = StringVar()

        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1000)
        self.rowconfigure(1, weight=1000)
        self.columnconfigure(1, weight=1)
        self.input_file = FileChooser(self, (('RWA files', '*.rwa'), ('All files', '*.*')),
            relief=tk.RIDGE, borderwidth=2) # test border
        self.input_file.grid(row=0, column=0, columnspan=2, sticky=tk.W+tk.E+tk.N)
        self.editor = TkContourEditor(self)
        self.editor.get_tk_widget().grid(row=1, column=0, sticky=tk.W+tk.E+tk.N+tk.S)
        self.side_panel = tk.Frame(self, relief=tk.RIDGE, borderwidth=2)
        self.side_panel.grid(row=1, column=1, sticky=tk.E+tk.N+tk.S, padx=2, pady=2)

        r = c = 1
        self.analysis_label = tk.Label(self.side_panel, text='Analysis:')
        self.analysis_label.grid(column=c, row=r, sticky=tk.W)
        self.analysis_input = ttk.Combobox(self.side_panel, state='readonly',
            textvariable=self.analysis)
        self.analysis_input.grid(column=c+1, row=r, sticky=tk.W)
        r += 1
        self.cell_label = tk.Label(self.side_panel, text='Cell:')
        self.cell_label.grid(column=c, row=r, sticky=tk.W)
        self.cell_input = tk.Entry(self.side_panel, width=4, textvariable=self.editor.cell)
        self.cell_input.grid(column=c+1, row=r, sticky=tk.W)
        r += 1
        self.step_label = tk.Label(self.side_panel, text='Step:')
        self.step_label.grid(column=c, row=r, sticky=tk.W)
        self.step_input = tk.Spinbox(self.side_panel, width=2, from_=1, to=99, increment=1,
            textvariable=self.editor.step)
        self.step_input.grid(column=c+1, row=r, sticky=tk.W)
        r += 1
        self.variable_label = tk.Label(self.side_panel, text='Variable:')
        self.variable_label.grid(column=c, row=r, sticky=tk.W)
        self.variable_input = ttk.Combobox(self.side_panel, state='readonly',
            textvariable=self.editor.variable)
        self.variable_input.grid(column=c+1, row=r, sticky=tk.W)
        r += 1
        self.integral_label = tk.Label(self.side_panel, text='Sum:')
        self.integral_label.grid(column=c, row=r, sticky=tk.W)
        self.integral_output = tk.Label(self.side_panel, textvariable=self.editor.integral)
        self.integral_output.grid(column=c+1, row=r, sticky=tk.W)
        r += 1
        self.delaunay_button = tk.Button(self.side_panel, text='Toggle Delaunay',
            command=self.editor.toggle_delaunay)
        self.delaunay_button.grid(column=c, columnspan=2, row=r, sticky=tk.W)
        #for _r in range(r):
        #       self.side_panel.rowconfigure(_r, weight=1)
        for _c in range(c):
            self.side_panel.columnconfigure(_c, weight=1)
        for child in self.side_panel.winfo_children():
            child.grid_configure(padx=2, pady=2)
        self.disable_widgets()


    def new_file(self, *args, **kwargs):
        self.analysis.set('')
        self.analyses = None
        self.map_labels = None
        self.disable_widgets()
        map_file = self.input_file.filepath.get()
        try:
            self.analyses = load_rwa(map_file)
        except KeyError:
            try:
                # old format
                store = HDF5Store(map_file, 'r')
                store.lazy = False
                maps = peek_maps(store, store.store)
            finally:
                store.close()
        if self.analyses is None:
            try:
                tess_file = maps.rwa_file
            except AttributeError:
                # even older
                tess_file = maps.imt_file
            maps = maps.maps
            if not isinstance(tess_file, str):
                tess_file = tess_file.decode('utf-8')
            tess_file = os.path.join(os.path.dirname(map_file), tess_file)
            try:
                store = HDF5Store(tess_file, 'r')
                store.lazy = False
                try:
                    cells = store.peek('cells')
                except:
                    raise
                else:
                    if cells.tessellation is None:
                        cells._tessellation = store.peek('_tesselation', store.store['cells'])
                    self.editor.cells = cells
                finally:
                    store.close()
            except:
                print(traceback.format_exc())
                print('cells not found')
            self.set_maps(maps)
        else:
            self.map_labels = [ ' - '.join(str(a) for a in path)
                for path in label_paths(self.analyses, Maps) ]
            self.analysis_input['values'] = self.map_labels
            if not self.map_labels:
                raise ValueError('no analyses found')
            if not self.map_labels[1:]:
                self.analysis.set(self.map_labels[0])
        self.enable_widgets()

    def disable_widgets(self):
        self.analysis_label['state'] = 'disabled'
        self.analysis_input.state(('disabled',))
        self.cell_label['state'] = 'disabled'
        self.cell_input['state'] = 'disabled'
        self.variable_label['state'] = 'disabled'
        self.variable_input.state(('disabled',))
        self.editor.cell.set(-1)
        self.editor.map = None

    def enable_widgets(self):
        if self.map_labels:
            self.analysis_label['state'] = 'normal'
            self.analysis_input.state(('!disabled',))
        if self.editor.cells:
            self.cell_label['state'] = 'normal'
            self.cell_input['state'] = 'normal'
        if self.editor.map is not None:
            self.variable_label['state'] = 'normal'
            self.variable_input.state(('!disabled',))

    @property
    def debug(self):
        return self.editor.debug

    @debug.setter
    def debug(self, d): # this setter is never called
        self.editor.debug = d

    def _analysis(self, *args):
        analysis = self.analysis.get()
        try:
            if not analysis:
                return
            labels = []
            for label in analysis.split(' - '):
                try:
                    label = int(label)
                except (TypeError, ValueError):
                    pass
                labels.append(label)
            cells, maps = find_artefacts(self.analyses, (Partition, Maps), labels)
            assert isinstance(maps, Maps)
            self.editor.cells = cells
            self.set_maps(maps)
            self.enable_widgets()
        except:
            print(traceback.format_exc())
            raise

    def set_maps(self, maps):
        if isinstance(maps, Maps):
            maps = maps.maps
        self.editor.map = maps
        self.variable_input['values'] = self.editor.variables
        if not self.editor.variables[1:]:
            self.editor.variable.set(self.editor.variables[0])


def contour_utility():
    root = tk.Tk()
    root.protocol('WM_DELETE_WINDOW', sys.exit)
    root.rowconfigure(0, weight=1)
    root.columnconfigure(0, weight=1)
    app = ContourEditingApp(root)
    app.grid(row=0, column=0, sticky=tk.W+tk.E+tk.N+tk.S)
    app.editor.debug = '-d' in sys.argv or '--debug' in sys.argv # app.debug property does not work
    app.mainloop()

if __name__ == '__main__':
    matplotlib.use('TkAgg')
    contour_utility()




# -*- coding: utf-8 -*-

# Copyright © 2017-2019, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


import numpy as np
from collections import namedtuple, OrderedDict
import numpy.ma as ma
from pandas import Series, DataFrame

Matrix = namedtuple('Matrix', 'size shape dtype order')

class ArrayChain(object):
    __slots__ = ('members', 'order')

    def __init__(self, *members, **kwargs):
        order = kwargs.pop('order', 'simple') # Py2 workaround
        if kwargs:
            raise TypeError('expected at most 1 keyword arguments, got {}'.format(len(kwargs)+1))
        self.order = order
        self.members = OrderedDict()
        if members:
            if not members[1:]:
                members = members[0]
            if not isinstance(members[0], tuple):
                members = zip(members[0::2], members[1::2])
        for member, example in members:
            rd = 'C'
            if isinstance(example, tuple):
                sz = np.prod(np.asarray(example[0]))
                try:
                    sh, dt, rd = example
                except:
                    sh, dt = example
            else:
                sz, sh, dt = example.size, example.shape, example.dtype
            self.members[member] = Matrix(sz, sh, dt, rd)

    def __contains__(self, member):
        return member in self.members

    def __getitem__(self, member):
        return self.members[member]

    def __setitem__(self, member, example):
        rd = 'C'
        if isinstance(example, tuple):
            sz = np.prod(np.asarray(example[0]))
            try:
                sh, dt, rd = example
            except:
                sh, dt = example
        else:
            sz, sh, dt = example.size, example.shape, example.dtype
        self.members[member] = Matrix(sz, sh, dt, rd)

    def __delitem__(self, member):
        del self.members[member]

    def _at(self, a, member, matrix):
        k = list(self.members.keys()).index(member)
        if self.order == 'simple':
            i0 = sum([ m.size for m in list(self.members.values())[:k] ])
            return slice(i0, i0 + matrix.size)
        else:
            raise NotImplementedError("only 'simple' order is supported")

    def at(self, a, member):
        matrix = self.members[member]
        return self._at(a, member, matrix)

    def get(self, a, member):
        matrix = self.members[member]
        return a[self._at(a, member, matrix)].reshape(matrix.shape, order=matrix.order)

    def set(self, a, member, m):
        if isinstance(m, (Series, DataFrame)):
            m = m.values
        matrix = self.members[member]
        if matrix.shape == m.shape:
            a[self._at(a, member, matrix)] = m.flatten(matrix.order)
        else:
            raise ValueError('wrong matrix size')

    @property
    def size(self):
        return sum([ m.size for m in self.members.values() ])

    @property
    def shape(self):
        return (self.size,)



class ChainArray(ArrayChain):
    __slots__ = ('combined',)

    def __init__(self, *members, **kwargs):
        ArrayChain.__init__(self, *members, **kwargs)
        self.combined = np.empty(self.shape)
        # adapted copy-paste from ArrayChain.__init__
        if not members[1:]:
            members = members[0]
        kwargs = dict(zip(members[0::2], members[1::2]))
        #
        if any([isinstance(v, ma.MaskedArray) for v in kwargs.values()]):
            self.combined = ma.asarray(self.combined)
        for k in kwargs:
            self[k] = kwargs[k]

    def __getitem__(self, k):
        return self.get(self.combined, k)

    def __setitem__(self, k, v):
        self.set(self.combined, k, v)

    def update(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError('numpy array expected')
        if isinstance(self.combined, ma.MaskedArray):
            self.combined = ma.array(x, mask=self.combined.mask)
        else:
            self.combined = x


__all__ = [
    'Matrix',
    'ArrayChain',
    'ChainArray',
    ]




# -*- coding: utf-8 -*-

# Copyright © 2020, Institut Pasteur
#   Contributor: François Laurent

# This file is part of the TRamWAy software available at
# "https://github.com/DecBayComp/TRamWAy" and is distributed under
# the terms of the CeCILL license as circulated at the following URL
# "http://www.cecill.info/licenses.en.html".

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


from .abc import *
from .proxy import *
from tramway.tessellation import plugins


class TessellerPlugin(TessellerProxy):
    """
    Wraps any plugin referenced in `tramway.tessellation.plugins`.
    """
    __slots__ = ('_module','_setup')
    def __init__(self, name, **kwargs):
        try:
            setup, module = plugins[name]
        except KeyError:
            raise KeyError("no such tesseller plugin: '{}'".format(name))
        cls = setup['make']
        if isinstance(cls, str):
            cls = getattr(module, cls)
        TessellerProxy.__init__(self, cls, **kwargs)
        self._reset_kwargs()
        self._module, self._setup = module, setup
    def _reset_kwargs(self):
        TessellerProxy._reset_kwargs(self)
        self._init_kwargs = { attr: None for attr in \
                ('ref_distance',
                 'min_distance','avg_distance','max_distance',
                 'min_probability','avg_probability','max_probability',
                ) }


def tesseller_plugin(name):
    """
    Translate plugin names into `RWAnalyzer`-ready tessellers as
    defined in the *tessellers* module, if available,
    else returns a generic `TessellerPlugin` initializer.
    """
    from . import proxied
    if name == 'grid':
        return proxied.Squares
    elif name == 'hexagon':
        return proxied.Hexagons
    elif name == 'kmeans':
        return proxied.KMeans
    elif name in ('gwr', 'gas'):
        return proxied.GWR
    else:
        class Cls(TessellerPlugin):
            __slots__ = () # important for __setattr__ to work properly!
            def __init__(self, **kwargs):
                TessellerPlugin.__init__(self, name, **kwargs)
        return Cls

__all__ = ['TessellerPlugin', 'tesseller_plugin']

